
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from .transformer import Transformer1, Transformer2, TransformerWithGlobalCond, Transformer1WithAttn, Transformer2WithAttn

from nflows.flows import Flow
from nflows.distributions import StandardNormal, ConditionalDiagonalNormal
from nflows.transforms import CompositeTransform, RandomPermutation, AffineCouplingTransform, ActNorm, PiecewiseRationalQuadraticCDF, AffineTransform
from nflows.nn.nets import ResidualNet
from nflows.utils import torchutils

from torch.distributions import Normal, Categorical, MixtureSameFamily, Independent


def transformer_nf_model(args, **kwargs):
    
    ### Transformer model ###
    if args.model_name == "transformer1":
        model_class = Transformer1
    elif args.model_name == "transformer2":
        model_class = Transformer2
    elif args.model_name == "transformer1_with_global_cond":
        model_class = TransformerWithGlobalCond
        transformer_cls = Transformer1
    elif args.model_name == "transformer2_with_global_cond":
        model_class = TransformerWithGlobalCond
        transformer_cls = Transformer2
    elif args.model_name == "transformer1_with_attn":
        model_class = Transformer1WithAttn
    elif args.model_name == "transformer2_with_attn":
        model_class = Transformer2WithAttn
    else:
        raise ValueError(f"Invalid model: {args.model_name}")
    
    common_args = dict(
            num_condition = args.num_features_cond, 
            d_model = args.d_model, 
            num_layers = args.num_layers, 
            num_heads = args.num_heads, 
            max_length = args.max_length, 
            num_features_in = args.num_features_in, 
            num_features_out = args.num_context, 
            num_token_types = 1,
            last_activation = nn.Tanh(), 
            pred_prob = False, 
            **kwargs
        )

    if "with_global_cond" in args.model_name:
        common_args["num_features_global"] = args.num_features_global
        common_args["transformer_cls"] = transformer_cls
    
    model = model_class(**common_args)

    ### Flow model ###
    transforms = []

    alpha = 0.05 # value in RealNVP
    transforms.append(
    AffineTransform(
        shift=torch.full([args.num_features_in], fill_value=alpha),
        scale=torch.full([args.num_features_in], fill_value=1.0 - 2.0*alpha),
    )
    )
    transforms.append(Logit())

    for ilayer in range(args.num_flows):
        #transforms.append(ActNorm(args.num_features_out)) # for stabilyzing training and quick convergence
        transforms.append(
            AffineCouplingTransform(
                #mask=torch.arange(args.num_features_out) % 2,
                mask = torch.arange(args.num_features_in) % 2 if ilayer % 2 == 0 else (torch.arange(args.num_features_in) + 1) % 2,
                transform_net_create_fn=lambda in_features, out_features: ResidualNet(
                    in_features=in_features,
                    out_features=out_features,
                    hidden_features=args.hidden_dim,
                    context_features=args.num_context,
                    num_blocks=2,
                    activation=nn.ReLU()
                )
            )
        )
        transforms.append(PiecewiseRationalQuadraticCDF(shape=[args.num_features_in], num_bins=8, tails='linear', tail_bound=3.0))
        transforms.append(RandomPermutation(args.num_features_in))
        
    transform = CompositeTransform(transforms)
    
    if args.base_dist == "normal":
        base_dist = StandardNormal(shape=[args.num_features_in])
    elif args.base_dist == "conditional_normal":
        base_dist = ConditionalDiagonalNormal(shape=[args.num_features_in], 
            context_encoder=nn.Sequential(nn.Linear(args.num_context, args.hidden_dim), 
                                          nn.LeakyReLU(), 
                                          nn.Linear(args.hidden_dim, 2 * args.num_features_in)
            )
        )

    elif args.base_dist == "bimodal":
        base_dist = BimodalNormal(shape=[args.num_features_in], offset=2.0)
    elif args.base_dist == "conditional_bimodal":
        base_dist = ConditionalBimodal(context_dim=args.num_context, latent_dim=args.num_features_in, hidden_dim=args.hidden_dim)
    else:
        raise ValueError(f"Invalid base distribution: {args.base_dist}")
    flow = Flow(transform, base_dist)

    return model, flow
    


def calculate_transformer_nf_loss(transformer, flow, batch, stop=None, stop_predictor=None):
    device = next(transformer.parameters()).device

    condition = batch["context"].to(device)
    seq = batch["target"].to(device)
    mask = batch["mask"].to(device) # (batch, max_length, num_context)
    global_cond = batch["global_context"].to(device)

    input_seq = seq[:, :-1] # (batch, max_length-1, num_features)
    output = transformer(condition, input_seq, global_cond=global_cond) # (batch, max_length, num_context)

    mask = mask[:,:,0].reshape(-1) # (batch * max_length)

    _, _, num_features = seq.shape
    _, _, num_context = output.shape

    context_for_flow = output.reshape(-1, num_context)[mask]  # flow context (num_galaxies, num_context)
    target_for_flow = seq.reshape(-1, num_features)[mask] # flow target (num_galaxies, num_features)
    log_prob = flow.log_prob(target_for_flow, context=context_for_flow)
    loss = - log_prob.mean() / num_features

    if stop is not None and stop_predictor is not None:
        stop = stop.to(device) # (batch, max_length)
        stop = stop.reshape(-1)[mask] # (num_galaxies,)
        stop_pred = stop_predictor(context_for_flow)
        loss_stop = torch.nn.BCELoss()(stop_pred, stop)  # (num_galaxies,)
        return loss, loss_stop
    else:
        return loss

def generate_with_transformer_nf(transformer, flow, x_cond, stop_predictor=None, prob_threshold=0, stop_threshold=None):
    transformer.eval()
    flow.eval()
    if stop_predictor is not None:
        stop_predictor.eval()

    batch_size = len(x_cond)
    max_length = transformer.max_length
    num_features = transformer.num_features_in

    x_seq = torch.zeros(batch_size, max_length, num_features).to(x_cond.device)
    stop_flags = torch.zeros(batch_size, dtype=torch.bool).to(x_cond.device)

    for t in range(max_length):
        with torch.no_grad():
            context = transformer(x_cond, x_seq[:,:t,:]) # (batch, t+1, num_context)
            context = context[:, -1, :] # (batch, num_context)
            if prob_threshold > 0:
                raise ValueError("# prob_threshold > 0 is not yet implemented")
                num_samples = 1
                embedded_context = flow._embedding_net(context)

                noise, log_prob = flow._distribution.sample_and_log_prob(num_samples, context)
                
                if embedded_context is not None:
                    # Merge the context dimension with sample dimension in order to apply the transform.
                    noise = torchutils.merge_leading_dims(noise, num_dims=2)
                    embedded_context = torchutils.repeat_rows(embedded_context, num_reps=1)

                samples, _ = flow._transform.inverse(noise, context=embedded_context)

                if embedded_context is not None:
                    # Split the context dimension from sample dimension.
                    samples = torchutils.split_leading_dim(samples, shape=[-1, num_samples])
                
            else:
                samples = flow.sample(1, context=context) # (batch, 1, num_features)

            x_seq[:, t, :] = samples.squeeze(1)
               
            if stop_predictor is not None:  
                stop_prob = stop_predictor(context) # (batch, )
                stop_now = stop_prob > stop_threshold
                stop_flags = stop_now
            
            elif stop_threshold is not None:
                if t > 0:
                    stop_now = x_seq[:,t,0] < stop_threshold # (batch, )
                    stop_flags |= stop_now

            if t == 0:
                if num_features > 1:
                    x_seq[:,0,1] = torch.randn(batch_size).to(x_cond.device) * 1e-3 # Random distance for central
                if num_features > 2:
                    x_seq[:,0,2] = torch.randn(batch_size).to(x_cond.device) * 1e-3
            
            if stop_flags.all():
                break

    return x_seq

def my_stop_predictor(args):
    return StopPredictor(context_dim=args.num_context, hidden_dim=args.hidden_dim_stop)


class StopPredictor(nn.Module):
    def __init__(self, context_dim, hidden_dim=64):
        super().__init__()
        self.layers = nn.Sequential(
            nn.Linear(context_dim, hidden_dim),
            nn.LeakyReLU(),
            nn.Linear(hidden_dim, 1),
            nn.Sigmoid()
        )
    
    def forward(self, context):
        x = self.layers(context)
        return x.squeeze(-1)  # (batch_size,) in [0, 1]

class BimodalNormal(StandardNormal):
    """A simple mixture of two Gaussians (bimodal) with fixed means."""

    def __init__(self, shape, offset=2.0):
        super().__init__(shape)
        self.offset = offset

    def _sample(self, num_samples, context=None):
        if context is None:
            device = self._log_z.device
            half = num_samples // 2

            samples1 = torch.randn(half, *self._shape, device=device) - self.offset
            samples2 = torch.randn(num_samples - half, *self._shape, device=device) + self.offset
            samples = torch.cat([samples1, samples2], dim=0)
            return samples[torch.randperm(num_samples)]  

        else:
            context_size = context.shape[0]
            total = context_size * num_samples
            device = context.device
            half = total // 2

            samples1 = torch.randn(half, *self._shape, device=device) - self.offset
            samples2 = torch.randn(total - half, *self._shape, device=device) + self.offset
            samples = torch.cat([samples1, samples2], dim=0)
            samples = samples[torch.randperm(total)]

            return torchutils.split_leading_dim(samples, [context_size, num_samples])

    def _log_prob(self, inputs, context=None):
        if inputs.shape[1:] != self._shape:
            raise ValueError(
                "Expected input of shape {}, got {}".format(
                    self._shape, inputs.shape[1:]
                )
            )

        x = inputs.view(inputs.size(0), -1)

        logp1 = -0.5 * torch.sum((x + self.offset)**2, dim=1) - self._log_z
        logp2 = -0.5 * torch.sum((x - self.offset)**2, dim=1) - self._log_z

        log_prob = torch.logaddexp(logp1, logp2) - np.log(2.0)
        return log_prob

    def _mean(self, context):
        if context is None:
            return self._log_z.new_zeros(self._shape)
        else:
            return context.new_zeros(context.shape[0], *self._shape)

class ConditionalBimodal(nn.Module):
    def __init__(self, context_dim, latent_dim, hidden_dim=64):
        super().__init__()
        self.latent_dim = latent_dim

        # context → offset (1,), logit(weight) (1,), scale (1,)
        self.net = nn.Sequential(
            nn.Linear(context_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, 2 * latent_dim + 2)  # mean1, mean2, scale, weight_logits
        )

    def get_distribution(self, context):
        batch_size = context.shape[0]
        params = self.net(context)  # (batch, 2*latent_dim + 2)

        # 分解
        mean1 = params[:, :self.latent_dim]
        mean2 = params[:, self.latent_dim:2*self.latent_dim]
        scale_raw = params[:, 2*self.latent_dim:2*self.latent_dim+1]  # shape (batch, 1)
        weight_logits = params[:, 2*self.latent_dim+1:]  # shape (batch, 1)

        scale = F.softplus(scale_raw) + 1e-3

        mix_logits = torch.cat([weight_logits, -weight_logits], dim=1)  # (batch, 2)
        mix_dist = Categorical(logits=mix_logits)

        means = torch.stack([mean1, mean2], dim=1)  # (batch, 2, latent_dim)
        scales = scale.unsqueeze(1).expand_as(means)  # (batch, 2, latent_dim)

        comp_dist = Independent(Normal(means, scales), 1)  # treat latent_dim as event dim
        mixture = MixtureSameFamily(mix_dist, comp_dist)
        return mixture

    def sample(self, num_samples, context):
        """
        context: (batch, context_dim)
        return: (batch, num_samples, latent_dim)
        """
        dist = self.get_distribution(context)  
        samples = dist.sample((num_samples,))  # (num_samples, batch, latent_dim)
        return samples.permute(1, 0, 2)  # (batch, num_samples, latent_dim)

    def log_prob(self, x, context):
        """
        x: (batch, latent_dim)
        context: (batch, context_dim)
        return: (batch,)
        """
        dist = self.get_distribution(context)
        return dist.log_prob(x)
    

from nflows.transforms.base import InputOutsideDomain, Transform, InverseTransform
    
class Sigmoid(Transform):
    def __init__(self, temperature=1, eps=1e-6, learn_temperature=False):
        super().__init__()
        self.eps = eps
        if learn_temperature:
            self.temperature = nn.Parameter(torch.Tensor([temperature]))
        else:
            temperature = torch.Tensor([temperature])
            self.register_buffer('temperature', temperature)

    def forward(self, inputs, context=None):
        temperature = self.temperature.to(inputs.device)

        inputs = temperature * inputs
        outputs = torch.sigmoid(inputs)

        logabsdet = torchutils.sum_except_batch(
            torch.log(temperature) - F.softplus(-inputs) - F.softplus(inputs)
        )
        return outputs, logabsdet

    def inverse(self, inputs, context=None):
        if torch.min(inputs) < 0 or torch.max(inputs) > 1:
            raise InputOutsideDomain()

        inputs = torch.clamp(inputs, self.eps, 1 - self.eps)

        temperature = self.temperature.to(inputs.device)
        outputs = (1 / temperature) * (torch.log(inputs) - torch.log1p(-inputs))
        logabsdet = -torchutils.sum_except_batch(
            torch.log(temperature)
            - F.softplus(-temperature * outputs)
            - F.softplus(temperature * outputs)
        )
        return outputs, logabsdet


class Logit(InverseTransform):
    def __init__(self, temperature=1, eps=1e-6):
        super().__init__(Sigmoid(temperature=temperature, eps=eps))


import os
import sys
import argparse
import json
import copy
import h5py
import re

from tqdm import tqdm

import numpy as np

import torch

cspeed = 3e10  # [cm/s]

from .io_utils import normalize, namespace_to_dict

def create_mask(array, threshold): 
    """
    mask out galaxies satisfying either of the following:
    - if it is central and the value is below threshold
    - if it is satellite and any of the satellites before it is below threshold
    """

    _, seq_length = array.shape

    mask_valid = array > threshold # (num_halos, max_length)
    mask_below = array <= threshold # (num_halos, max_length)
    mask_below[:, 0] = False  

    first_below = np.where(mask_below.any(axis=1), mask_below.argmax(axis=1), seq_length)
    indices = np.arange(seq_length)[None, :]  # (1, max_length)
    mask = indices < first_below[:, None]  # (num_halos, max_length)

    mask = mask & mask_valid # (num_halos, max_length)

    return mask

def generate_galaxy(args, x_in, global_params=None, verbose=True):
    """
    args: args.gpu_id, args.model_dir, args.threshold, and args.max_sfr_file are used
    x_in: (num_halos, num_features_in); halo properties
    """

    print("# Use Transformer to generate SFR")

    from cosmoglint.model.transformer import transformer_model
    device = torch.device("cuda:{}".format(args.gpu_id) if torch.cuda.is_available() else "cpu")

    ### load Transformer
    with open("{}/args.json".format(args.model_dir), "r") as f:
        opt = json.load(f, object_hook=lambda d: argparse.Namespace(**d))
        opt.norm_param_dict = namespace_to_dict(opt.norm_param_dict)

    model = transformer_model(opt)
    model.load_state_dict(torch.load("{}/model.pth".format(args.model_dir), map_location="cpu"))
    model.to(device)
    model.eval()
    
    if verbose:
        print("opt: ", opt)
        print(model)

    ### generate galaxies
    print("# Generate galaxies (batch size: {:d})".format(opt.batch_size))
    for i, key in enumerate(opt.input_features):
        x_in[...,i] = normalize(x_in[...,i], key, opt.norm_param_dict)
    x_in = torch.from_numpy(x_in).float().to(device)

    if global_params is not None:
        global_params = global_params[opt.global_features].to_numpy(dtype=np.float32)
        for i, key in enumerate(opt.global_features):
            global_params[...,i] = normalize(global_params[...,i], key, opt.norm_param_dict)
        global_params = torch.tensor(np.array(global_params), dtype=torch.float32).to(device)

    if args.max_sfr_file is None:
        print("# No max SFR file provided, using default max IDs")
        max_ids = None
    else:
        max_ids = np.loadtxt(args.max_sfr_file)
        max_ids = ( max_ids * opt.num_features_out ).astype(int)
        max_ids = torch.tensor(max_ids).to(device) # (num_features, )
    
    num_batch = (len(x_in) + opt.batch_size - 1) // opt.batch_size
    stop_criterion = normalize(args.threshold, opt.output_features[0], opt.norm_param_dict) # stop criterion for SFR
    generated = []
    for batch_idx in tqdm(range(num_batch)):
        start = batch_idx * opt.batch_size 
        x_batch = x_in[start: start + opt.batch_size] # (batch_size, num_features)
        global_cond_batch = global_params.unsqueeze(0).repeat(len(x_batch), 1) if global_params is not None else None # (batch_size, num_global_features)
        with torch.no_grad():
            generated_batch, _ = model.generate(x_batch, global_cond=global_cond_batch, prob_threshold=1e-5, stop_criterion=stop_criterion, max_ids=max_ids) # (batch_size, seq_length, num_features)
            
        generated.append(generated_batch.cpu().detach().numpy())
        
    generated = np.concatenate(generated, axis=0) # (num_halos, seq_length, num_features) or (num_halos, seq_length * num_features, 1)

    if opt.use_flat_representation:
        generated = generated.squeeze(-1).reshape(len(generated), -1, opt.num_features_in) # (num_halos, max_length, num_features) 
        mask = mask.reshape(len(mask), -1, opt.num_features_in)

    mask = create_mask(generated[:,:,0], stop_criterion) # (num_halos, seq_length)

    # De-normalize
    for i, key in enumerate(opt.output_features):
        generated[...,i] = normalize(generated[...,i], key, opt.norm_param_dict, inverse=True)

    print("# Number of valid galaxies: {:d}".format(len(generated)))
    
    return generated, mask

def generate_galaxy_TransNF(args, x_in, global_params=None, verbose=True):
    """
    args: args.gpu_id, args.model_dir, and args.threshold are used
    x_in: (num_halos, num_features_in), halo properties
    """

    print("# Use Transformer-NF to generate galaxies")

    from cosmoglint.model.transformer_nf import transformer_nf_model, generate_with_transformer_nf
    device = torch.device("cuda:{}".format(args.gpu_id) if torch.cuda.is_available() else "cpu")
    print("Using device: ", device)

    ### load Transformer
    with open("{}/args.json".format(args.model_dir), "r") as f:
        opt = json.load(f, object_hook=lambda d: argparse.Namespace(**d))
        opt.norm_param_dict = namespace_to_dict(opt.norm_param_dict)

    model, flow = transformer_nf_model(opt)

    model.load_state_dict(torch.load("{}/model.pth".format(args.model_dir), map_location="cpu"))
    model.to(device)
    model.eval()
    
    flow.load_state_dict(torch.load("{}/flow.pth".format(args.model_dir), map_location="cpu"))
    flow.to(device)
    flow.eval()

    if verbose:
        print("opt: ", opt)
        print(model)
        print(flow)

    ### generate galaxies
    print("# Generate galaxies (batch size: {:d})".format(opt.batch_size))
    
    for i, key in enumerate(opt.input_features):
        x_in[...,i] = normalize(x_in[...,i], key, opt.norm_param_dict)
    x_in = torch.from_numpy(x_in).float().to(device)

    if global_params is not None:
        global_params = np.array([global_params[name] for name in opt.global_features], dtype=np.float32)
        for i, key in enumerate(opt.global_features):
            global_params[...,i] = normalize(global_params[...,i], key, opt.norm_param_dict)
        global_params = torch.from_numpy(global_params).float().to(device)
    
    num_batch = (len(x_in) + opt.batch_size - 1) // opt.batch_size
    generated = []
    def stop_criterion(sample):
        # sample: (batch, num_features)
        return (sample[:, 0] < 1).all()
    
    for batch_idx in tqdm(range(num_batch)):
        start = batch_idx * opt.batch_size 
        x_batch = x_in[start: start + opt.batch_size] # (batch_size, 1)
        global_cond_batch = global_params.unsqueeze(0).repeat(len(x_batch), 1) if global_params is not None else None
        generated_batch = generate_with_transformer_nf(model, flow, x_batch, global_cond=global_cond_batch, stop_criterion=stop_criterion) # (batch_size, max_length, num_features)
        generated.append(generated_batch.cpu().detach().numpy())
    generated = torch.cat(generated, dim=0) # (num_halos, max_length, num_features) or (num_halos, max_length * num_features, 1)
     
    # De-normalize
    for i, key in enumerate(opt.output_features):
        generated[...,i] = normalize(generated[...,i], key, opt.norm_param_dict, inverse=True)

    # Set mask for selection
    sfr = generated[...,0]
    mask = create_mask(sfr, args.threshold) # (num_halos, seq_length)   
    
    print("# Number of valid galaxies: {:d}".format(len(generated)))
    
    return generated, mask

def populate_galaxies_in_cube(args, x_in, pos, vel, redshift, cosmo, global_params=None):
    
    if "Transformer_NF" in args.model_dir:
        generated, mask = generate_galaxy_TransNF(args, x_in, global_params=global_params)
    else:
        generated, mask = generate_galaxy(args, x_in, global_params=global_params)

    seq_length = mask.shape[1]
    num_features = generated.shape[-1]
    num_gal = mask.sum()

    # Define flag_central
    flag_central = np.zeros_like(mask, dtype=bool) # (num_halos, seq_length)
    flag_central[:, 0] = True

    # Flatten the arrays
    mask = mask.reshape(-1) # (num_halos * seq_length, )
    generated = generated.reshape(-1, num_features) # (num_halos * seq_length, num_features)
    pos_central = np.repeat(pos[:,None,:], seq_length, axis=1).reshape(-1, 3) # (num_halos * seq_length, 3)
    vel_central = np.repeat(vel[:,None,:], seq_length, axis=1).reshape(-1, 3) # (num_halos * seq_length, 3)
    flag_central = flag_central.reshape(-1) # (num_halos * seq_length, )

    # Apply mask to arrays
    generated = generated[mask] # (num_galaxies_valid, num_features)
    pos_central = pos_central[mask] # (num_galaxies_valid, 3)
    vel_central = vel_central[mask] # (num_galaxies_valid, 3)
    flag_central = flag_central[mask] # (num_galaxies_valid, )

    # Distribute galaxies in cube
    print("# Generate positions of galaxies")

    sfr = generated[:,0]
    distance = generated[:,1]

    phi = np.random.uniform(0, 2 * np.pi, size=num_gal)
    cos_theta = np.random.uniform(-1, 1, size=num_gal)
    sin_theta = np.sqrt(1 - cos_theta ** 2)    

    pos_galaxies = pos_central
    pos_galaxies[:,0] += distance * sin_theta * np.cos(phi)
    pos_galaxies[:,1] += distance * sin_theta * np.sin(phi)
    pos_galaxies[:,2] += distance * cos_theta

    pos_galaxies_real = copy.deepcopy(pos_galaxies) 

    # Add redshift-space distortion
    if args.redshift_space:
        import astropy.units as u
        H = cosmo.H(redshift).to(u.km/u.s/u.Mpc).value #[km/s/Mpc]
        hlittle = cosmo.H(0).to(u.km/u.s/u.Mpc).value / 100.0 
        scale_factor = 1 / (1 + redshift)

        relative_vel_rad = generated[:,2]
        relative_vel_tan = generated[:,3]
        relative_vel_rad[flag_central] = 0 # Set vr to 0 for central galaxies
        alpha = np.random.uniform(0, 2 * np.pi, size=num_gal)
        vz_gal = - relative_vel_rad * cos_theta + relative_vel_tan * sin_theta * np.cos(alpha)
        pos_galaxies[:,2] += ( vel_central[:,2] + vz_gal )/ scale_factor / H * hlittle

    return sfr, pos_galaxies_real, pos_galaxies

def populate_galaxies_in_lightcone(args, x_in, pos, redshift, cosmo, global_params=None):
    """
    args: args.gpu_id, args.model_dir, args.model_config_file, args.args.threshold, and args.param_dir are used
    x_in: (num_halos, )
    pos: (num_halos, 3)
    redshift: (num_halos, )
    """

    print("# Use Transformer to generate SFR")

    opt = copy.deepcopy(args)

    generated_all = []
    pos_central_all = [] 
    redshift_central_all = [] 
    flag_central_all = []

    with open(args.model_config_file, "r") as f:
        snapshot_dict_str = json.load(f)
        snapshot_dict = {int(k): v for k, v in snapshot_dict_str.items()}

    print("# Model config:", snapshot_dict)
    redshifts_of_snapshots = np.array([ v[1] for v in snapshot_dict.values() ])    
    bin_edges = (redshifts_of_snapshots[:-1] + redshifts_of_snapshots[1:]) / 2.0
    bin_indices = np.digitize(redshift, bin_edges)  

    if args.param_dir is None:
        max_sfr_file_list = [ None for snapshot_number in snapshot_dict ]
    else:
        max_sfr_file_list = ["{}/max_nbin20_{:d}.txt".format(args.param_dir, snapshot_number) for snapshot_number in snapshot_dict]

    for i, snapshot_number in enumerate(snapshot_dict):
        model_path, redshift_of_snapshot = snapshot_dict[snapshot_number]
        print("# Snapshot number: {:d}, Redshift: {:.2f}".format(snapshot_number, redshift_of_snapshot))
        
        ### Skip if no haloes in this redshift bin
        mask_z = (bin_indices == i)
        if not np.any(mask_z):
            print("# No haloes in redshift bin {:d} (snapshot number {:d}), skipping...".format(i, snapshot_number))
            continue
    
        x_now = x_in[mask_z] # (num_halos_in_bin, )
        pos_now = pos[mask_z] # (num_halos_in_bin, 3)
        redshift_now = redshift[mask_z] # (num_halos_in_bin, 1)

        opt.model_dir = "{}/{}".format(args.model_dir, model_path)
        opt.max_sfr_file = max_sfr_file_list[i]

        if "Transformer_NF" in opt.model_dir:
            generated, mask = generate_galaxy_TransNF(opt, x_now, global_params=global_params, verbose=False)
        else:
            generated, mask = generate_galaxy(opt, x_now, global_params=global_params, verbose=False)
            
        seq_length = mask.shape[1]
        num_features = generated.shape[-1]

        # Define flag_central
        flag_central = np.zeros_like(mask, dtype=bool)
        flag_central[:, 0] = True

        # Flatten the arrays
        mask = mask.reshape(-1)
        generated = generated.reshape(-1, num_features) # (num_halos * seq_length, num_features)
        pos_central = np.repeat(pos_now[:,None,:], seq_length, axis=1).reshape(-1, 3) # (num_halos * seq_length, 3)
        redshift_central = np.repeat(redshift_now[:,None], seq_length, axis=1).reshape(-1) # (num_halos * seq_length)
        flag_central = flag_central.reshape(-1)

        # Apply mask to arrays
        generated = generated[mask] # (num_galaxies_valid, num_features)
        pos_central = pos_central[mask] # (num_galaxies_valid, 3)
        redshift_central = redshift_central[mask] # (num_galaxies_valid, 3)
        flag_central = flag_central[mask] # (num_galaxies_valid, )
        
        # Append
        generated_all.append(generated)
        pos_central_all.append(pos_central)
        redshift_central_all.append(redshift_central)
        flag_central_all.append(flag_central)

    generated_all = np.concatenate(generated_all, axis=0) # (num_galaxies_valid, num_features)
    pos_central_all = np.concatenate(pos_central_all, axis=0)
    redshift_central_all = np.concatenate(redshift_central_all, axis=0) # (num_galaxies_valid,)
    flag_central_all = np.concatenate(flag_central_all, axis=0) # (num_galaxies_valid,)
    
    ### Distribute galaxies in lightcone
    sfr = generated_all[:,0]
    distance = generated_all[:,1]

    num_gal = len(sfr)

    # Determine positions of galaxies
    print("# Generate positions of galaxies")
    _phi = np.random.uniform(0, 2 * np.pi, size=num_gal)
    _cos_theta = np.random.uniform(-1, 1, size=num_gal)
    _sin_theta = np.sqrt(1 - _cos_theta ** 2)
    
    # Convert Mpc to deg
    from .cosmology_utils import cMpc_to_arcsec, dcMpc_to_dz
    distance_arcsec = cMpc_to_arcsec(distance, redshift_central_all, cosmo=cosmo, l_with_hlittle=True)
    distance_z = dcMpc_to_dz(distance, redshift_central_all, cosmo=cosmo, l_with_hlittle=True)

    pos_galaxies = pos_central_all
    pos_galaxies[:,0] += distance_arcsec * _sin_theta * np.cos(_phi)
    pos_galaxies[:,1] += distance_arcsec * _sin_theta * np.sin(_phi)
    pos_galaxies[:,2] += distance_z * _cos_theta
    
    # Add redshift-space distortion
    if args.redshift_space:

        relative_vel_rad = generated_all[:,2]
        relative_vel_tan = generated_all[:,3]
        relative_vel_rad[flag_central_all] = 0 # Set vr to 0 for central galaxies
        alpha = np.random.uniform(0, 2 * np.pi, size=num_gal)
        vz_gal = - relative_vel_rad * _cos_theta + relative_vel_tan * _sin_theta * np.cos(alpha)
        
        beta = vz_gal / (cspeed * 100) # [(km/s) / (km/s)]

        redshift_rest = pos_galaxies[:,2]
        pos_galaxies[:,2] = ( 1. + redshift_rest ) * np.sqrt( (1. + beta) / (1. - beta) ) - 1.0


    return sfr, pos_galaxies, redshift_central_all



import random

import torch
import torch.nn as nn

from torch.distributions import Categorical

def transformer_model(args, **kwargs):
    
    if "transformer" in args.model_name:
        if args.model_name == "transformer1":
            model_class = Transformer1
        elif args.model_name == "transformer2":
            model_class = Transformer2
        elif args.model_name == "transformer1_with_global_cond":
            model_class = TransformerWithGlobalCond
            transformer_cls = Transformer1
        elif args.model_name == "transformer2_with_global_cond":
            model_class = TransformerWithGlobalCond
            transformer_cls = Transformer2
        elif args.model_name == "transformer1_with_attn":
            model_class = Transformer1WithAttn
        elif args.model_name == "transformer2_with_attn":
            model_class = Transformer2WithAttn
        else:
            raise ValueError(f"Invalid model: {args.model_name}")

        if len(args.output_features) != args.num_features_in:
            raise ValueError(f"num_features ({args.num_features_in}) is not consistent with the list of output features ({args.output_features})")      
          
        common_args = dict(
            d_model=args.d_model,
            num_layers=args.num_layers,
            num_heads=args.num_heads,
            num_condition=args.num_features_cond,
            num_features_out=args.num_features_out,
            output_features=args.output_features,
            **kwargs,
        )

        if args.use_flat_representation:
            common_args["max_length"] = args.max_length * args.num_features_in
            common_args["num_features_in"] = 1
            common_args["num_token_types"] = args.num_features_in
        
        else:
            common_args["max_length"] = args.max_length 
            common_args["num_features_in"] = args.num_features_in
            common_args["num_token_types"] = 1
        
        
        if "with_global_cond" in args.model_name:
            common_args["num_features_global"] = args.num_features_global
            common_args["transformer_cls"] = transformer_cls
        
        model = model_class(**common_args)

    else:
        raise ValueError(f"Invalid model: {args.model}")

    return model

class TransformerBase(nn.Module):
    def __init__(
            self, 
            num_condition = 1, 
            d_model = 128, 
            num_layers = 4, 
            num_heads = 8, 
            max_length = 10, 
            num_features_in = 1, 
            num_features_out = 1, 
            num_token_types = 1, 
            output_features = ["SubhaloSFR"], 
            central_values = {"SubhaloDist": 0.0, "SubhaloVrad": 0.5}, 
            dropout = 0
        ):
        super().__init__()

        self.d_model = d_model
        self.max_length = max_length
        self.num_features_in = num_features_in
        self.num_features_out = num_features_out
        self.num_token_types = num_token_types

        self.output_idx_map = {name: i for i, name in enumerate(output_features)}
        self.output_features = output_features
        self.central_values = central_values

        # Pisition and feature type embedding
        actual_max_length = max_length // num_token_types
        self.pos_embedding = nn.Embedding(actual_max_length, d_model)
        self.token_type_embedding = nn.Embedding(num_token_types, d_model)

        token_pos_id = torch.arange(actual_max_length).repeat_interleave(num_token_types)
        token_type_id = torch.arange(num_token_types).repeat(actual_max_length)
        self.register_buffer("token_pos_id", token_pos_id.long())
        self.register_buffer("token_type_id", token_type_id.long())

    def forward(self, context, x, global_cond=None):
        raise NotImplementedError("forward method not implemented")

    def generate_square_subsequent_mask(self, sz):
        mask = torch.triu(torch.ones(sz, sz), diagonal=1)
        mask = mask.masked_fill(mask==1, float('-inf'))
        return mask
    
    def _set_to_zero(self, x, mask):
        zero_tensor = torch.tensor(0.0).to(x.device)
        return torch.where(mask, zero_tensor, x)
        
    def generate(
            self, 
            context, 
            global_cond = None, 
            seq = None, 
            teacher_forcing_ratio = 0.0, 
            temperature = 1.0, 
            stop_criterion = None, 
            prob_threshold = 1e-5, 
            monotonicity_start_index = 1, 
            max_ids = None, 
            buffer_percent = 0.05
        ):
        """
        context: (B, num_features_in)
        global_cond: (B, num_features_global) -- simply ignored
        seq: (B, L, num_features_out) 
        teacher_forcing_ratio: 
        monotonicity_start_index: from which galaxy to enforce monotonicity. No enforcement if < 0.
        max_ids: torch tensor listing the indices of maximum value for the primary parameter for differen
        """
        # context: (batch, num_condition)
        batch_size = len(context)

        if len(context.shape) == 1:
            context = context.unsqueeze(-1)

        # used for enforce_monotonicity and max_ids
        # buffer = 1 indicates the bins just above the max_ids are avoided.
        buffer = max(int(buffer_percent * self.num_features_out), 1)
        if max_ids is not None:
            max_ids = max_ids.to(context.device) + buffer # (nbins, ) 
            nbins = len(max_ids)
            context_bins = torch.linspace(0, 1, nbins, device=context.device) # (nbins, )
            context_bin_indices = torch.bucketize(context[:, 0], context_bins) - 1 # (batch, )
            bin_indices = torch.arange(self.num_features_out, device=context.device) # (num_features_out, )
            mask_max_ids = (bin_indices.unsqueeze(0) > max_ids[context_bin_indices].unsqueeze(1)) # (batch, num_features_out)

        generated = torch.zeros(batch_size, self.max_length, self.num_features_in).to(context.device) # (batch, max_length, num_features_in)
        mask_all_batch = torch.ones(batch_size, dtype=torch.bool).to(context.device)
                    
        for t in range(self.max_length):

            if seq is not None and t < seq.size(1) and random.random() < teacher_forcing_ratio:
                next_token = seq[:, t]
            else:
                x = self(context, generated[:, :t], global_cond=global_cond) 
                # generated[:, :t]: (batch, t, num_features_in)
                # x: (batch, t+1, num_faetures_in, num_features_out)
                
                x_last = x[:, -1, :, :] 
                # last taken (batch, num_features_in, num_features_out)
            
                x_last = x_last / temperature

                token_type = t % self.num_token_types

                if token_type == 0:
                    if monotonicity_start_index is not None:
                        # Set the probability at x(t) >= x(t-1) to zero for the primary parameter
                        if t > monotonicity_start_index: 
                            previous_token_bin = (generated[:, t - self.num_token_types, 0] * self.num_features_out).long() + buffer
                            previous_token_bin = previous_token_bin.contiguous().view(-1, 1) # (batch, 1)
                            bin_indices = torch.arange(self.num_features_out, device=context.device).view(1, -1) # (1, num_features_out)
                            mask = (bin_indices > previous_token_bin) # (batch, num_features_out)
                            mask = mask & (x_last[:, 0, :] >= prob_threshold) # (batch, num_features_out)
                            x_last[:, 0, :] = self._set_to_zero(x_last[:, 0, :], mask) # set the probability to zero for bins above the previous bin

                    if max_ids is not None:
                        # Even when monotonicity is not enforced, set the probability at x > x_max to zero for the primary parameter if max_ids is defined.
                        x_last[:, 0, :] = self._set_to_zero(x_last[:, 0, :], mask_max_ids)

                x_last = self._set_to_zero(x_last, x_last < prob_threshold) # set the probability to zero if less than prob_threshold

                x_last = x_last.reshape(-1, self.num_features_out) # (batch * num_features_in, num_features_out)
                bin_indices = Categorical(probs=x_last).sample().float().view(-1, self.num_features_in) # (batch, num_features_in)
                uniform_noise = torch.rand_like(bin_indices, device=context.device)  # (batch, num_features_in)
                next_token = (bin_indices + uniform_noise) / self.num_features_out  # (batch, num_features_in)

                if token_type == 0:
                    next_token[:, 0] = self._set_to_zero(next_token[:, 0], next_token[:,0] < 1./ self.num_features_out) # strictly set the sampled primary parameter to zero if it is less than 1/num_features_out 

                mask_all_batch = torch.ones(batch_size, dtype=torch.bool).to(context.device)
                
                # Set the central galaxy's parameters to fixed values
                is_first_gal = ( t // self.num_token_types == 0 )
                if is_first_gal:
                    if self.num_token_types == 1:
                        for feat, cval in self.central_values.items():
                            idx = self.output_idx_map.get(feat)
                            if idx is not None:
                                next_token[:, idx] = cval + self._set_to_zero(next_token[:, idx], mask_all_batch)
                    else:
                        feat = self.output_features[token_type]
                        cval = self.central_values.get(feat)
                        if cval is not None:
                            next_token[:, 0] = cval + self._set_to_zero(next_token[:, 0], mask_all_batch)

            # Stop generation if the primary parameter is below criterion
            if token_type == 0:
                if stop_criterion is not None:            
                    if torch.all(next_token[:,0] < stop_criterion):
                        return generated, x

            generated[:, t, :] = next_token # (batch, num_features_in)           
            
        if seq is not None and teacher_forcing_ratio > 0:
            x = self(context, generated[:,:-1])

        return generated, x


class Transformer1(TransformerBase): # add logM at first in the sequence
    def __init__(
            self, 
            num_condition = 1, 
            d_model = 128, 
            num_layers = 4, 
            num_heads = 8, 
            max_length = 10, 
            num_features_in = 1, 
            num_features_out = 1, 
            num_token_types = 1, 
            output_features = ["SubhaloSFR"], 
            dropout = 0, 
            last_activation = nn.Softmax(dim=-1), 
            pred_prob = True    
        ):
        super().__init__(num_condition=num_condition, d_model=d_model, num_layers=num_layers, num_heads=num_heads, max_length=max_length, num_features_in=num_features_in, num_features_out=num_features_out, num_token_types=num_token_types, output_features=output_features, dropout=dropout)

        self.embedding_layers = nn.Sequential(
            nn.Linear(num_features_in, d_model),
            nn.LeakyReLU(),
            nn.Dropout(dropout),
            nn.Linear(d_model, d_model),
        ) 
        self.context_embedding_layers = nn.Sequential(
            nn.Linear(num_condition, d_model),
            nn.LeakyReLU(),
            nn.Dropout(dropout),
            nn.Linear(d_model, d_model),
        )

        decoder_layer = nn.TransformerDecoderLayer(d_model=d_model, nhead=num_heads, batch_first=True, dropout=dropout)
        self.decoder = nn.TransformerDecoder(decoder_layer, num_layers=num_layers)

        self.pred_prob = pred_prob
        if pred_prob:
            self.output_layer = nn.Linear(d_model, num_features_in*num_features_out) 
        else:
            self.output_layer = nn.Linear(d_model, num_features_out)

        self.out_activation = last_activation
        

    def forward(self, context, x, global_cond=None):
        # context: (batch, num_condition)
        # x: (batch, seq_length, num_features_in)
        
        batch_size, seq_length, num_features_in = x.shape  
        total_seq_length = 1 + seq_length # total length of (context and x)

        # concatenate embeddings of context and x 
        context = context.view(batch_size, 1, -1) # (batch, 1, num_condition)
        for layer in self.context_embedding_layers:
            context = layer(context)  
        # context: (batch, 1, d_model)

        for layer in self.embedding_layers:
            x = layer(x)
        x = torch.cat([context, x], dim=1)  # (batch, seq_length + 1, d_model)
        
        # add position and type embedding
        pos_emb = self.pos_embedding(self.token_pos_id[:total_seq_length]).unsqueeze(0) # (1, seq_length + 1, d_model)
        type_emb = self.token_type_embedding(self.token_type_id[:total_seq_length]).unsqueeze(0) # (1, seq_length + 1, d_model)
        x = x + pos_emb + type_emb # (batch, seq_length + 1, d_model)

        # decode
        causal_mask = self.generate_square_subsequent_mask(total_seq_length).to(x.device)
        dummy_memory = torch.zeros(batch_size, 1, self.d_model, device=x.device)
        x = self.decoder(x, memory=dummy_memory, tgt_mask=causal_mask)  # (batch, seq_length + 1, d_model)
    
        # output layer
        x = self.output_layer(x)  # (batch, seq_length + 1, num_features_in * num_features_out) or (batch, seq_length + 1, num_features_out)
        
        if self.pred_prob:
            x = x.view(batch_size, total_seq_length, self.num_features_in, -1) # (batch, seq_length + 1, num_features_in, num_features_out)

        x = self.out_activation(x)
        # x = entmax(x, dim=-1)

        return x

class Transformer2(TransformerBase): # embed context and x together, and then add positional embedding
    def __init__(
            self, 
            num_condition = 1, 
            d_model = 128, 
            num_layers = 4, 
            num_heads = 8, 
            max_length = 10, 
            num_features_in = 1, 
            num_features_out = 1, 
            num_token_types = 1, 
            output_features = ["SubhaloSFR"], 
            dropout = 0, 
            last_activation = nn.Softmax(dim=-1), 
            pred_prob = True
        ):

        super().__init__(num_condition=num_condition, d_model=d_model, num_layers=num_layers, num_heads=num_heads, max_length=max_length, num_features_in=num_features_in, num_features_out=num_features_out, num_token_types=num_token_types, output_features=output_features, dropout=dropout)
        
        self.start_token = torch.ones(1) 
                
        self.embedding_layers = nn.Sequential(
            nn.Linear(num_condition+num_features_in, d_model),
            nn.LeakyReLU(),
            nn.Dropout(dropout),
            nn.Linear(d_model, d_model),
        )

        decoder_layer = nn.TransformerDecoderLayer(d_model=d_model, nhead=num_heads, batch_first=True, dropout=dropout)
        self.decoder = nn.TransformerDecoder(decoder_layer, num_layers=num_layers)

        self.pred_prob = pred_prob
        if pred_prob:
            self.output_layer = nn.Linear(d_model, num_features_in*num_features_out)
        else:
            self.output_layer = nn.Linear(d_model, num_features_out)

        self.out_activation = last_activation
    
    def forward(self, context, x, global_cond=None):
        # context: (batch, num_condition)
        # x: (batch, seq_length, num_features_in)

        batch_size, seq_length, num_features_in = x.shape 
        total_seq_length = 1 + seq_length # total length of (start token and x)

        context = context.view(batch_size, 1, -1) # (batch, 1, num_condition)
        context = context.expand(batch_size, seq_length+1, -1) # (batch, seq_length+1, num_condition)

        ## add start token
        start_token = self.start_token.expand(batch_size, 1, self.num_features_in).to(x.device) # (batch, 1, num_features_in)
        x = torch.cat([start_token, x], dim=1) # (batch, seq_length+1, num_features_in) 

        ## concatenate context and x
        x = torch.cat([context, x], dim=2)  # (batch, seq_length+1, num_condition + num_features_in)

        ## embedding
        for layer in self.embedding_layers:
            x = layer(x)  
        # x: (batch, seq_length+1, d_model)

        # add position and type embedding
        pos_emb = self.pos_embedding(self.token_pos_id[:total_seq_length]).unsqueeze(0) # (1, seq_length + 1, d_model)
        type_emb = self.token_type_embedding(self.token_type_id[:total_seq_length]).unsqueeze(0) # (1, seq_length + 1, d_model)
        x = x + pos_emb + type_emb # (batch, seq_length + 1, d_model)
        
        # decode
        causal_mask = self.generate_square_subsequent_mask(seq_length+1).to(x.device)
        dummy_memory = torch.zeros(batch_size, 1, self.d_model, device=x.device)
        x = self.decoder(x, memory=dummy_memory, tgt_mask=causal_mask)  # (batch, seq_length+1, d_model)
    
        # output layer
        x = self.output_layer(x)  # (batch, seq_length+1, num_features_in * num_features_out) or (batch, seq_length+1, num_features_out)

        if self.pred_prob:
            x = x.view(batch_size, seq_length+1, self.num_features_in, -1) # (batch, seq_length+1, num_features_in, num_features_out)
        
        x = self.out_activation(x)

        return x


class TransformerWithGlobalCond(nn.Module): 
    def __init__(
            self, 
            num_features_global, 
            transformer_cls = Transformer1,
            num_condition = 1, 
            d_model = 128, 
            num_layers = 4, 
            num_heads = 8, 
            max_length = 10, 
            num_features_in = 1, 
            num_features_out = 1, 
            num_token_types = 1, 
            output_features = ["SubhaloSFR"], 
            dropout = 0, 
            last_activation = nn.Softmax(dim=-1), 
            pred_prob = True
        ):
        super().__init__()

        self.transformer = transformer_cls(
            num_condition = num_condition + num_features_global,
            d_model = d_model, 
            num_layers = num_layers, 
            num_heads = num_heads, 
            max_length = max_length, 
            num_features_in = num_features_in, 
            num_features_out = num_features_out, 
            num_token_types = num_token_types, 
            output_features = output_features,
            dropout = dropout, 
            last_activation = last_activation, 
            pred_prob = pred_prob
        )
        
    def forward(self, context, x, global_cond):
        if len(context.shape) == 1:
            context = context.unsqueeze(-1)
        ctx = torch.cat([context, global_cond], dim=1)
        return self.transformer(ctx, x)
    
    def generate(self, context, global_cond, **kwargs):
        if len(context.shape) == 1:
            context = context.unsqueeze(-1)
        ctx = torch.cat([context, global_cond], dim=1)
        return self.transformer.generate(ctx, **kwargs)

    
from typing import Optional

class TransformerDecoderLayerWithAttn(nn.TransformerDecoderLayer):
    def _sa_block(
            self, 
            x: torch.Tensor, 
            attn_mask: Optional[torch.Tensor],
            key_padding_mask: Optional[torch.Tensor], 
            is_causal: bool = False
        ) -> torch.Tensor:
    
        attn_output, attn_weights = self.self_attn(
            x, x, x,
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
            need_weights=True,
            is_causal=is_causal,
        )

        self.attn_weights = attn_weights.detach().cpu()
        return self.dropout1(attn_output)
    

class Transformer1WithAttn(Transformer1):
    def __init__(
            self, 
            num_condition = 1, 
            d_model = 128, 
            num_layers = 4, 
            num_heads = 8, 
            max_length = 10, 
            num_features_in = 1, 
            num_features_out = 1, 
            num_token_types = 1, 
            output_features = ["SubhaloSFR"], 
            dropout = 0, 
            last_activation = nn.Softmax(dim=-1), 
            pred_prob = True
        ):
        super().__init__(num_condition=num_condition, d_model=d_model, num_layers=num_layers, num_heads=num_heads, max_length=max_length, num_features_in=num_features_in, num_features_out=num_features_out, num_token_types=num_token_types, output_features=output_features, dropout=dropout, last_activation=last_activation, pred_prob=pred_prob)
        
        decoder_layer = TransformerDecoderLayerWithAttn(d_model=d_model, nhead=num_heads, batch_first=True, dropout=dropout)
        self.decoder = nn.TransformerDecoder(decoder_layer, num_layers=num_layers)

class Transformer2WithAttn(Transformer2):
    def __init__(
            self, 
            num_condition = 1, 
            d_model = 128, 
            num_layers = 4, 
            num_heads = 8, 
            max_length = 10, 
            num_features_in = 1, 
            num_features_out = 1, 
            num_token_types = 1, 
            output_features = ["SubhaloSFR"], 
            dropout = 0, 
            last_activation = nn.Softmax(dim=-1), 
            pred_prob = True
        ):
        super().__init__(num_condition=num_condition, d_model=d_model, num_layers=num_layers, num_heads=num_heads, max_length=max_length, num_features_in=num_features_in, num_features_out=num_features_out, num_token_types=num_token_types, output_features=output_features, dropout=dropout, last_activation=last_activation, pred_prob=pred_prob)

        decoder_layer = TransformerDecoderLayerWithAttn(d_model=d_model, nhead=num_heads, batch_first=True, dropout=dropout)
        self.decoder = nn.TransformerDecoder(decoder_layer, num_layers=num_layers)





from .generation_utils import generate_galaxy, generate_galaxy_TransNF, populate_galaxies_in_lightcone
from .io_utils import MyDataset, load_global_params, load_halo_data, load_lightcone_data, my_save_model


import os
import sys
import argparse
import json

from tqdm import tqdm
import numpy as np
import torch

import warnings, os

import copy

def short_formatwarning(msg, category, filename, lineno, line=None):
    return f"{os.path.basename(filename)}:{lineno}: {category.__name__}: {msg}\n"

warnings.formatwarning = short_formatwarning
warnings.filterwarnings("always", category=RuntimeWarning)

import astropy.units as u
from astropy.constants import c as cspeed # [m/s]
from astropy.cosmology import FlatLambdaCDM
cosmo_default = FlatLambdaCDM(H0=67.74, Om0=0.3089)

def cMpc_to_arcsec(l_cMpc, z, cosmo=cosmo_default, l_with_hlittle=False): 
    if l_with_hlittle:
        hlittle = cosmo.H0.value / 100
        l_cMpc = l_cMpc / hlittle # [Mpc/h] -> [Mpc]
    l_rad = l_cMpc * u.Mpc / cosmo.comoving_transverse_distance(z)
    l_arcsec = (l_rad * u.radian).to(u.arcsec)
    return l_arcsec.value

def arcsec_to_cMpc(l_arcsec, z, cosmo=cosmo_default, l_with_hlittle=False):
    l_rad = l_arcsec * u.arcsec / u.radian
    l_cMpc = ( cosmo.comoving_transverse_distance(z) * l_rad ).to(u.Mpc)
    if l_with_hlittle:
        hlittle = cosmo.H0.value / 100
        l_cMpc = l_cMpc * hlittle # [Mpc] -> [Mpc/h]
    return l_cMpc.value 

def dcMpc_to_dz(l_cMpc, z, cosmo=cosmo_default, l_with_hlittle=False):
    if l_with_hlittle:
        hlittle = cosmo.H0.value / 100
        l_cMpc = l_cMpc / hlittle # [Mpc/h] -> [Mpc]
    dx_dz = (cspeed  / cosmo.H(z)).to(u.Mpc)
    d_z = l_cMpc / dx_dz.value 
    return d_z

def dz_to_dcMpc(dz, z, cosmo=cosmo_default, l_with_hlittle=False):
    dx_dz = (cspeed / cosmo.H(z)).to(u.Mpc)
    l_cMpc = dz * dx_dz.value 
    if l_with_hlittle:
        hlittle = cosmo.H0.value / 100
        l_cMpc = l_cMpc * hlittle # [Mpc] -> [Mpc/h]
    return l_cMpc # [Mpc/h] if l_with_hlittle else [Mpc]

def freq_to_comdis(nu_obs, nu_rest, cosmo=cosmo_default, l_with_hlittle=False):
    z = nu_rest / nu_obs - 1
    if z < 0:
        print("Error: z < 0")
        sys.exit(1)
    l_cMpc = cosmo.comoving_distance(z).to(u.Mpc).value
    if l_with_hlittle:
        hlittle = cosmo.H0.value / 100
        l_cMpc = l_cMpc * hlittle # [Mpc] -> [Mpc/h]

    return l_cMpc # [Mpc/h] if l_with_hlittle else [Mpc]

def z_to_log_lumi_dis(z, cosmo=cosmo_default):
    return np.log10( cosmo.luminosity_distance(z).to(u.cm).value )

def populate_galaxies_in_lightcone(args, logm, pos, redshift, cosmo=cosmo_default):
    """
    args: args.gpu_id, args.model_dir, args.model_config_file, args.args.threshold, and args.param_dir are used
    logm: (num_halos, )
    pos: (num_halos, 3)
    redshift: (num_halos, )
    """

    print("# Use Transformer to generate SFR")

    opt = copy.deepcopy(args)

    generated_all = []
    pos_central_all = [] 
    redshift_central_all = [] 
    flag_central_all = []

    with open(args.model_config_file, "r") as f:
        snapshot_dict_str = json.load(f)
        snapshot_dict = {int(k): v for k, v in snapshot_dict_str.items()}

    print("# Model config:", snapshot_dict)
    redshifts_of_snapshots = np.array([ v[1] for v in snapshot_dict.values() ])    
    bin_edges = (redshifts_of_snapshots[:-1] + redshifts_of_snapshots[1:]) / 2.0
    bin_indices = np.digitize(redshift, bin_edges)  

    if args.param_dir is None:
        max_sfr_file_list = [ None for snapshot_number in snapshot_dict ]
    else:
        max_sfr_file_list = ["{}/max_nbin20_{:d}.txt".format(args.param_dir, snapshot_number) for snapshot_number in snapshot_dict]

    for i, snapshot_number in enumerate(snapshot_dict):
        model_path, redshift_of_snapshot = snapshot_dict[snapshot_number]
        print("# Snapshot number: {:d}, Redshift: {:.2f}".format(snapshot_number, redshift_of_snapshot))
        
        ### Skip if no haloes in this redshift bin
        mask_z = (bin_indices == i)
        if not np.any(mask_z):
            print("# No haloes in redshift bin {:d} (snapshot number {:d}), skipping...".format(i, snapshot_number))
            continue
    
        logm_now = logm[mask_z] # (num_halos_in_bin, )
        pos_now = pos[mask_z] # (num_halos_in_bin, 3)
        redshift_now = redshift[mask_z] # (num_halos_in_bin, 1)

        opt.model_dir = "{}/{}".format(args.model_dir, model_path)
        opt.max_sfr_file = max_sfr_file_list[i]
        opt.prob_threshold = 1e-5

        if "Transformer_NF" in opt.model_dir:
            raise ValueError("Transformer_NF model is not supported yet. Please use a different model.")
        else:
            from .generation_utils import generate_galaxy
            generated, mask = generate_galaxy(opt, logm_now)
            
        seq_length = mask.shape[1]
        num_features = generated.shape[-1]

        # Define flag_central
        flag_central = np.zeros_like(mask, dtype=bool)
        flag_central[:, 0] = True

        # Flatten the arrays
        mask = mask.reshape(-1)
        generated = generated.reshape(-1, num_features) # (num_halos * seq_length, num_features)
        pos_central = np.repeat(pos_now[:,None,:], seq_length, axis=1).reshape(-1, 3) # (num_halos * seq_length, 3)
        redshift_central = np.repeat(redshift_now[:,None], seq_length, axis=1).reshape(-1) # (num_halos * seq_length, 3)
        flag_central = flag_central.reshape(-1)

        # Apply mask to arrays
        generated = generated[mask] # (num_galaxies_valid, num_features)
        pos_central = pos_central[mask] # (num_galaxies_valid, 3)
        redshift_central = redshift_central[mask] # (num_galaxies_valid, 3)
        flag_central = flag_central[mask] # (num_galaxies_valid, )
        
        # Append
        generated_all.append(generated)
        pos_central_all.append(pos_central)
        redshift_central_all.append(redshift_central)
        flag_central_all.append(flag_central)

    generated_all = np.concatenate(generated_all, axis=0) # (num_galaxies_valid, num_features)
    pos_central_all = np.concatenate(pos_central_all, axis=0)
    redshift_central_all = np.concatenate(redshift_central_all, axis=0) # (num_galaxies_valid,)
    flag_central_all = np.concatenate(flag_central_all, axis=0) # (num_galaxies_valid,)
    
    ### Distributes galaxies in lightcone
    sfr = generated_all[:,0]
    distance = generated_all[:,1]

    num_gal = len(sfr)

    # Determine positions of galaxies
    print("# Generate positions of galaxies")
    _phi = np.random.uniform(0, 2 * np.pi, size=num_gal)
    _cos_theta = np.random.uniform(-1, 1, size=num_gal)
    _sin_theta = np.sqrt(1 - _cos_theta ** 2)
    
    # Convert Mpc to deg
    distance_arcsec = cMpc_to_arcsec(distance, redshift_central_all, cosmo=cosmo, l_with_hlittle=True)
    distance_z = dcMpc_to_dz(distance, redshift_central_all, cosmo=cosmo, l_with_hlittle=True)

    pos_galaxies = pos_central
    pos_galaxies[:,0] += distance_arcsec * _sin_theta * np.cos(_phi)
    pos_galaxies[:,1] += distance_arcsec * _sin_theta * np.sin(_phi)
    pos_galaxies[:,2] += distance_z * _cos_theta
    
    # Add redshift-space distortion
    if args.redshift_space:

        relative_vel_rad = generated[:,2]
        relative_vel_tan = generated[:,3]
        relative_vel_rad[flag_central] = 0 # Set vr to 0 for central galaxies
        alpha = np.random.uniform(0, 2 * np.pi, size=num_gal)
        vz_gal = - relative_vel_rad * _cos_theta + relative_vel_tan * _sin_theta * np.cos(alpha)
        
        beta = vz_gal / (cspeed * 100) # [(km/s) / (km/s)]

        redshift_rest = pos_galaxies[:,2]
        pos_galaxies[:,2] = ( 1. + redshift_rest ) * np.sqrt( (1. + beta) / (1. - beta) ) - 1.0


    return sfr, pos_galaxies, redshift_central_all



import sys
import os
from argparse import Namespace
import random
import numpy as np

import h5py

import torch

from tqdm import tqdm

from torch.utils.data import Dataset


def my_save_model(model, fname):
    torch.save(model.state_dict(), fname)
    print(f"# Model saved to {fname}")

def save_catalog_data(pos_list, value, args, output_fname):
    if not isinstance(pos_list, list):
        pos_list = [pos_list]

    with open(output_fname, 'w') as f:
        for i, v in enumerate(value):
            f.write(f"{pos_list[0][i, 0]} {pos_list[0][i, 1]} ")
            for pos in pos_list:
                f.write(f"{pos[i, 2]} ")

            f.write(f"{v}\n")

    print(f"# Catalog saved to {output_fname}")

def save_intensity_data(intensity, args, output_fname):
    args_dict = vars(args)
    args_dict = {k: (v if v is not None else "None") for k, v in args_dict.items()}

    if not isinstance(intensity, list):
        intensity = [intensity]
    
    with h5py.File(output_fname, 'w') as f:
        for i, d in enumerate(intensity):
            f.create_dataset(f'intensity{i}', data=d)    
        for key, value in args_dict.items():
            f.attrs[key] = value
    print(f"# Data cube saved as {output_fname}")

def namespace_to_dict(ns):
    if isinstance(ns, Namespace):
        return {k: namespace_to_dict(v) for k, v in vars(ns).items()}
    elif isinstance(ns, dict):
        return {k: namespace_to_dict(v) for k, v in ns.items()}
    else:
        return ns

def convert_to_log(val, val_min):
    log_val = np.full_like(val, val_min)
    mask = val > 10**val_min
    log_val[mask] = np.log10(val[mask])
    return log_val

def convert_to_log_with_sign(val):
    return np.sign(val) * np.log10(np.abs(val) + 1)

def inverse_convert_to_log_with_sign(val):
    return np.sign(val) * ( 10 ** np.abs( val ) - 1 )

def normalize(x, key, norm_param_dict, inverse=False, convert=True):
    """
    x: array-like
    key: str
    norm_param_dict: dict
        e.g., {
            "GroupMass": {"min": 10, "max": 15, "norm": "log"},
            ...
        }
    inverse: bool
        If True, perform inverse normalization.
    convert: bool
        If True, convert to/from log scale based on norm_param_dict.
    """

    x = np.array(x)
        
    if norm_param_dict is not None:
        xmin = norm_param_dict[key]["min"]
        xmax = norm_param_dict[key]["max"] 
        norm = norm_param_dict[key]["norm"]

        if inverse:
            x = x * ( xmax - xmin ) + xmin
            if convert:
                if norm == "log":
                    x = 10 ** x
                elif norm == "log_with_sign":
                    x = inverse_convert_to_log_with_sign(x)
        else:
            if convert:
                if norm == "log":
                    x = convert_to_log(x, xmin)    
                elif norm == "log_with_sign":
                    x = convert_to_log_with_sign(x)
            x = ( x - xmin ) / ( xmax - xmin )

    return x

def load_global_params(global_param_file, global_features, norm_param_dict=None):

    if global_features is None:
        global_params = None

    else:
        if global_param_file is None:
            raise ValueError("global_param_file must be specified when global_features is provided.")
        
        if not isinstance(global_param_file, list):
            global_param_file = [global_param_file]

        global_params = []
        for f in global_param_file:
            data = np.genfromtxt(f, names=True, dtype=None, encoding="utf-8")
            global_params_now = np.vstack([data[name] for name in global_features]).T.astype(np.float32)
            global_params.append(global_params_now)

        global_params = np.vstack(global_params)

        for i, key in enumerate(global_features):
            global_params[...,i] = normalize(global_params[...,i], key, norm_param_dict)        

    return global_params # (ndata, num_features_global)
    
def load_halo_data(
        file_path, 
        input_features,
        output_features,
        norm_param_dict=None, 
        max_length=10, 
        sort=True,
        ndata=None,
        exclude_ratio=0.0, 
        use_excluded_region=False,
    ):
        
    def load_values(f, key):
        if key not in f:
            raise ValueError(f"Key '{key}' not found in the file.")

        data = f[key][:]
        return data

    num_features_in = len(input_features)
    num_features_out = len(output_features)

    with h5py.File(file_path, "r") as f:

        # Load input features
        source_list = []
        for feature in input_features:
            x = load_values(f, feature)
            source_list.append(x)

        source = np.stack(source_list, axis=1)  # (N, num_features_in)
        for i, key in enumerate(input_features):
            source[:,i] = normalize(source[:,i], key, norm_param_dict)

        mask = np.ones(len(source), dtype=bool)
        for i in range(num_features_in):
            mask = mask & ( source[:,i] > 0 )
            
        if exclude_ratio > 0:
            boxsize = f.attrs["BoxSize"] # [kpc/h]
            halo_pos = f["GroupPos"][:]  # [kpc/h]
            mask_exclude = (halo_pos[:,0] > boxsize * (1.-exclude_ratio)) \
                        & (halo_pos[:,1] > boxsize * (1.-exclude_ratio)) \
                        & (halo_pos[:,2] > boxsize * (1.-exclude_ratio))

            if use_excluded_region:
                print("# Using excluded region of size ({} * BoxSize)^3".format(exclude_ratio))
                mask = mask & mask_exclude
            else:
                print("# Exclude halos in the corner of size ({} * BoxSize)^3".format(exclude_ratio))
                print("# The excluded region is {:.2f} % of the entire volume".format(100.0 * (exclude_ratio**3)))
                mask = mask & (~mask_exclude)
            
        if mask.sum() == 0:
            print("# No halo is found in {}".format(file_path))
            return torch.empty((0, num_features_in), dtype=torch.float32), []

        # Load output features
        target_list = []
        for feature in output_features:
            y = load_values(f, feature)
            target_list.append(y)

        target = np.stack(target_list, axis=1)  # (N, num_features_out)
        for i, key in enumerate(output_features):
            target[:,i] = normalize(target[:,i], key, norm_param_dict)

        num_subgroups = f["GroupNsubs"][:]

        offset = 0
        y_list = []
        for j in range(len(source)):
            start = offset
            end = start + num_subgroups[j]
            offset = end

            if not mask[j]:                
                continue

            if num_subgroups[j] == 0:
                y_j = np.zeros((1, num_features_out)) # handle empty subgroups
            else:
                y_j = target[start:end, :]

            if sort:
                sorted_indices = [0] + sorted(range(1, len(y_j)), key=lambda k: y_j[k,0], reverse=True)
                y_j = y_j[sorted_indices]

            y_j = y_j[:max_length] # truncate
            y_j = torch.tensor(y_j, dtype=torch.float32)
            y_list.append(y_j)
            
    x = source[mask]
    x = torch.tensor(x, dtype=torch.float32)
    
    if ndata is not None:
        x = x[:ndata]
        y_list = y_list[:ndata]

    return x, y_list

class MyDataset(Dataset):
    def __init__(
            self, 
            path,  
            input_features,
            output_features, 
            global_params=None,
            norm_param_dict=None, 
            max_length=10, 
            sort=True,
            ndata=None, 
            exclude_ratio=0.0,
            use_excluded_region=False,
            use_flat_representation=False,
            show_pbar=True,
        ):
            
        if not isinstance(path, list):
            path = [path]

        if global_params is not None:
            if len(global_params) != len(path):
                raise ValueError("The number of global parameter sets must match the number of data files")

        x = []
        self.y = []
        self.g = []

        plist = path
        if len(path) < 20:
            verbose = True 
        else:
            verbose = False
            if show_pbar:
                plist = tqdm(plist, file=sys.stderr)
            print("# Loading halo data from {} to {} ({} files)".format(path[0], path[-1], len(path)))

        for i, p in enumerate(plist):
            if verbose:
                print(f"# Loading halo data from {p}")
    
            x_tmp, y_tmp = load_halo_data(p, input_features, output_features, norm_param_dict=norm_param_dict, max_length=max_length, sort=sort, ndata=ndata, exclude_ratio=exclude_ratio, use_excluded_region=use_excluded_region)
            x.append(x_tmp) 
            self.y = self.y + y_tmp

            if global_params is not None:
                global_param = global_params[i]
                g_tmp = np.repeat(global_param[None, :], len(x_tmp), axis=0) # (Nhalo, num_features_global)
            else:
                g_tmp = np.zeros((len(x_tmp), 1)) # dummy (Nhalo, 1)
            
            self.g.append( g_tmp )

        self.x = torch.cat(x, dim=0)

        if len(self.x) == 0:
            raise ValueError("No halo is found.")

        self.g = np.vstack(self.g) 
        self.g = torch.tensor(self.g, dtype=torch.float32)
        
        _, num_params = (self.y[0]).shape

        self.y_padded = torch.zeros(len(self.x), max_length, num_params)
        self.mask = torch.zeros(len(self.x), max_length, num_params, dtype=torch.bool)
        
        for i, y_i in enumerate(self.y):
            length = len(y_i)
            self.y_padded[i, :length, :] = y_i[:max_length]
            self.mask[i, :length+1, :] = True # use the last + 1 value to learn when to stop
        
        if use_flat_representation:
            self.y_padded = self.y_padded.reshape(len(self.y_padded), -1, 1) # (Nhalo, max_length * output_features, 1)
            self.mask = self.mask.reshape(len(self.mask), -1, 1) # (Nhalo, max_length * output_features, 1)

    def __len__(self):
        return len(self.x)

    def __getitem__(self, idx):
        batch = {
            "context": self.x[idx],
            "global_context": self.g[idx],
            "target": self.y_padded[idx],
            "mask": self.mask[idx]
        }
        return batch
    
def load_lightcone_data(input_fname, cosmo):
    print(f"# Load {input_fname}")

    if "pinocchio" in input_fname: 
        if "old_version" in input_fname:
            M, theta, phi, _, redshift_obs, redshift_real = load_old_plc(input_fname)
            mass = M
        else:
            import cosmoglint.utils.ReadPinocchio5 as rp
            myplc = rp.plc(input_fname)
            
            mass = myplc.data["Mass"] 
            theta = myplc.data["theta"] # [arcsec]
            phi = myplc.data["phi"]

            redshift_obs = myplc.data["obsz"]
            redshift_real = myplc.data["truez"]

        import astropy.units as u
        hlittle = cosmo.H(0).to(u.km/u.s/u.Mpc).value / 100.0 
        mass /= hlittle # [Msun]

        theta = ( 90. - theta ) * 3600 # [arcsec]
        pos_x = theta * np.cos( phi * np.pi / 180. ) # [arcsec] 
        pos_y = theta * np.sin( phi * np.pi / 180. ) # [arcsec]
        
        print("# Minimum log mass in catalog: {:.5f}".format(np.min(np.log10(mass))))
        print("# Maximum pos: ({:.3f}, {:.3f}) arcsec".format(np.max(pos_x), np.max(pos_y)))
        print("# Minimum pos: ({:.3f}, {:.3f}) arcsec".format(np.min(pos_x), np.min(pos_y)))
        print("# Redshift: {:.3f} - {:.3f}".format(np.min(redshift_real), np.max(redshift_real)))
        print("# Number of halos: {}".format(len(mass)))

    else:
        raise ValueError("Unknown input file format")
    
    return mass, pos_x, pos_y, redshift_obs, redshift_real


def load_old_plc(filename):
    import struct

    plc_struct_format = "<Q d ddd ddd ddddd"  # Q=uint64, d=double, little-endian
    plc_size = struct.calcsize(plc_struct_format)

    M_list = []
    th_list = []
    ph_list = []
    vl_list = []
    zo_list = []
    z_list = []
    with open(filename, "rb") as f:
        while True:
            dummy_bytes = f.read(4)
            if not dummy_bytes:
                break  # EOF
            dummy = struct.unpack("<i", dummy_bytes)[0]

            plc_bytes = f.read(dummy)
            if len(plc_bytes) != dummy:
                break  # 不完全な読み込み

            data = struct.unpack(plc_struct_format, plc_bytes)
            (
                id, z, x1, x2, x3, v1, v2, v3,
                M, th, ph, vl, zo
            ) = data

            dummy2_bytes = f.read(4)
            dummy2 = struct.unpack("<i", dummy2_bytes)[0]

            M_list.append(M)
            th_list.append(th)
            ph_list.append(ph)
            vl_list.append(vl)
            zo_list.append(zo)
            z_list.append(z)
    
    return np.array(M_list), np.array(th_list), np.array(ph_list), np.array(vl_list), np.array(zo_list), np.array(z_list)



from .transformer import transformer_model, Transformer1, Transformer2, Transformer1WithAttn, Transformer2WithAttn
from .transformer_nf import transformer_nf_model, my_stop_predictor, calculate_transformer_nf_loss, generate_with_transformer_nf



# HEADER

"""
Routine for reading Pinocchio's binary catalogs and PLCs.
Usage:

# CATALOGS
import ReadPinocchio5 as rp
mycat = rp.catalog("pinocchio.0.0000.example.catalog.out")
print(mycat.Mass)

# PLC
import ReadPinocchio5 as rp
myplc = rp.plc("pinocchio.example.plc.out")
print(myplc.redshift)

# HISTORIES
import ReadPinocchio5 as rp
myhist = rp.histories("pinocchio.example.histories.out")
print(myhist.name)

Written by Pierluigi Monaco, Matteo Biagetti and Emiliano Munari

LAST MODIFIED:
Pierlugi Monaco 13/3/2025

"""


import numpy as np
import os
import sys
import copy
import struct


VERBOSE=False

class catalog:

    '''
    Reads a pinocchio catalog at fixed redshift
    Usage: mycat=rp.catalog(filename, silent=False, first_file=None, last_file=None)

    filename: name of catalog file, ending with .out. The code will recognise if the catalog is writen in several files
    silent: if True, the script does not give messages in the standard output
    first_file ,last_file: if the catalog if written in several files, it allows to select a limited range of output files

    Returns:
    mycat.data: the catalog, a structured numpy array
    mycat.cat_dtype: the data type of the catalog:
        [('name', numpy.int64),
        ('Mass', numpy.float32),
        ('pos', numpy.float32, 3),
        ('vel', numpy.float32, 3),
        ('posin', numpy.float32, 3),
        ('npart', numpy.int32)]       (absent in light format)
    mycat.Nhalos: number of halos in the catalog
    mycat.Nfiles:  number of files in which the catalog is written

    mycat.Mass: Mass field (for faster access)
    mycat.pos: pos field
    mycat.vel: vel field
    mycat.posin: posin field
    mycat.npart: npart field            (absent in light format)

    '''

    def __init__(self,filename,silent=False,first_file=None,last_file=None):

        if VERBOSE:
            silent=False

        # checks that the filename contains 'catalog'
        if not 'catalog' in filename:
            print("Are you sure you are providing the right file name?")
            if 'plc' in filename:
                print("...this looks like a plc file, please read it with rp.plc(filename)")
            elif 'histories' in filename:
                print("...this looks like a histories file, please read it with rp.histories(filename)")
            return None

        # checks that the input file ends by ".out"
        last_ext=filename.rfind('.')
        if filename[last_ext:]!='.out':
            print("The catalog file should end with .out, the file number extension will be checked by the code")
            return None

        # checks that the file exists, or that there are multiple files, and in case count them
        if not os.path.exists(filename):

            if not os.path.exists(filename+'.0'):

                print("file {} or {} not found:".format(filename,filename+'.0'))
                return None

            else:

                Nfiles=1
                while os.path.exists(filename+'.{}'.format(Nfiles)):
                    Nfiles+=1
                if not silent:
                    print("The catalog is written in {} files".format(Nfiles))

        else:

            Nfiles=1
            if not silent:
                print("The catalog is written in 1 file")

        # opens the (first) file and reads the record length
        if Nfiles==1:
            if not silent:
                print('reading header of file '+filename)
            reading = np.fromfile(filename,dtype=np.int32,count=10)
        else:
            if not silent:
                print('reading header of file '+filename+'.0')
            reading = np.fromfile(filename+'.0',dtype=np.int32,count=10)

        # number of tasks that write into a single file
        NTasksPerFile = reading[1]
        if not silent:
            print('This file has been written by {} tasks'.format(NTasksPerFile))

        # the header gives either the number of slices (always < 10) or the record length
        # in the first case the record length can be read from the 8th integer
        if reading[2]>10:
            newRun=True
            record_length=reading[2]
            Nslices=1
            if not silent:
                print('This is new output format, record length: {}'.format(record_length))
        else:
            newRun=False
            record_length=reading[7]
            Nslices=reading[2]
            if not silent:
                print('This is classic output format, record length: {}'.format(record_length))
                if Nslices==1:
                    print('The box has been fragmented in 1 slice')
                else:
                    print('The box has been fragmented in {} slices'.format(Nslices))


        # sets the record
        if record_length==96:          # this is the classic format in double precision

            self.cat_dtype=[ ('name',  np.int64),
                             ('Mass',  np.float64),
                             ('posin', np.float64, 3),
                             ('pos',   np.float64, 3),
                             ('vel',   np.float64, 3),
                             ('npart', np.int32) ]

            stored_dtype  =[ ('fort',  np.int32),
                             ('name',  np.int64),
                             ('Mass',  np.float64),
                             ('posin', np.float64, 3),
                             ('pos',   np.float64, 3),
                             ('vel',   np.float64, 3),
                             ('npart', np.int32),
                             ('pad'  , np.int32),
                             ('trof',  np.int32) ]


        elif record_length==56:

            if newRun:                     # the new format has posin in a different position

                self.cat_dtype=[ ('name',  np.int64),
                                 ('Mass',  np.float32),
                                 ('pos',   np.float32, 3),
                                 ('vel',   np.float32, 3),
                                 ('posin', np.float32, 3),
                                 ('npart', np.int32) ]

                stored_dtype  =[ ('name',  np.int64),
                                 ('Mass',  np.float32),
                                 ('pos',   np.float32, 3),
                                 ('vel',   np.float32, 3),
                                 ('posin', np.float32, 3),
                                 ('npart', np.int32),
                                 ('pad'  , np.int32) ]

            else:                          # this is the classic format in single precision

                self.cat_dtype=[ ('name',  np.int64),
                                 ('Mass',  np.float32),
                                 ('posin', np.float32, 3),
                                 ('pos',   np.float32, 3),
                                 ('vel',   np.float32, 3),
                                 ('npart', np.int32) ]

                stored_dtype  =[ ('fort',  np.int32),
                                 ('name',  np.int64),
                                 ('Mass',  np.float32),
                                 ('posin', np.float32, 3),
                                 ('pos',   np.float32, 3),
                                 ('vel',   np.float32, 3),
                                 ('npart', np.int32),
                                 ('pad'  , np.int32),
                                 ('trof',  np.int32) ]

        elif record_length==48:        # this is the new light format

            self.cat_dtype=[ ('name',  np.int64),
                             ('Mass',  np.float32),
                             ('pos',   np.float32, 3),
                             ('vel',   np.float32, 3),
                             ('posin', np.float32, 3) ]

            stored_dtype = self.cat_dtype


        elif record_length==40:        # this was used in NewClusterMocks

            self.cat_dtype=[ ('name',  np.int64),
                             ('Mass',  np.float32),
                             ('pos',   np.float32, 3),
                             ('vel',   np.float32, 3),
                             ('npart', np.int32) ]

            stored_dtype  =[ ('fort',  np.int32),
                             ('name',  np.int64),
                             ('Mass',  np.float32),
                             ('pos',   np.float32, 3),
                             ('vel',   np.float32, 3),
                             ('npart', np.int32),
                             ('trof',  np.int32) ]

        else:
            print("sorry, I do not recognize this record length")
            return None

        # decides what files to read
        if Nfiles>1:
            if first_file is None:
                first_file=0
            elif first_file<0:
                first_file=0
            elif first_file>Nfiles:
                first_file=Nfiles
            if last_file is None:
                last_file=Nfiles
            else:
                last_file += 1      
                if last_file<first_file:
                    last_file=first_file+1
                elif last_file>Nfiles:
                    last_file=Nfiles
            # this is to be used to define a pythonic range
            if not silent:
                print("I will read files in the python range from {} to {}".format(first_file,last_file))
        else:
            first_file=0
            last_file=Nfiles
        self.Nfiles=Nfiles

        # prepares to read the file(s)
        self.data=None
        NhalosPerFile=np.zeros(Nfiles,dtype=np.int64)

        # loops on the files to be read
        for myfile in range(first_file,last_file):

            if Nfiles==1:
                myfname = filename
            else:
                myfname = filename+'.{}'.format(myfile)

            if VERBOSE:
                print("reading file {}".format(myfname))

            # reads the file as a binary object
            with open(myfname,'rb') as f:
                bindata=f.read()
                f.close()
            filesize = len(bindata)

            # reads the number of halos contained in each header
            Nblocks=NTasksPerFile * Nslices     # this is the number of blocks to read
            Nwritten=0                          # some tasks may have no halos to write
            pos=16
            vechalo=[]
            while pos < filesize:
                # reads the number of halos to read
                vec=struct.Struct('iii').unpack(bindata[pos:pos+12])[1]
                pos+=12
                vechalo.append(vec)
                if vec>0:
                    if newRun:
                        pos += 8+vec*record_length
                    else:
                        pos += vec*(record_length+8)

            vechalo = np.asarray(vechalo)
            #print(vechalo.size, Nblocks, vechalo)
            NhalosPerFile[myfile]=vechalo.sum()
            Nwritten=(vechalo>0).sum()

            if VERBOSE:
                print(f"Found {Nwritten} non-void blocks in this file over {Nblocks}")

            # checks that the lenght of data is as expected
            if newRun:
                file_length = 16 + Nblocks*12 + Nwritten*8 + NhalosPerFile[myfile]*record_length
            else:
                file_length = 16 + Nblocks*12 + NhalosPerFile[myfile]*(record_length+8)

            if file_length != filesize:
                print(f'ERROR: inconsistency in the file size, should be {file_length} but I found {filesize} bytes')
                return None
            elif VERBOSE:
                print(f'predicted file length {file_length} matches with file size {filesize}')

            # format to select information from the byte object
            if newRun:
                cleanForm='16x '
                for block in range(Nblocks):
                    if vechalo[block]>0:
                        cleanForm+='16x {}s 4x '.format(vechalo[block]*record_length)
                    else:
                        cleanForm+='12x '

            else:
                cleanForm='16x '
                for block in range(Nblocks):
                    if vechalo[block]>0:
                        cleanForm+='12x {}s '.format(vechalo[block]*(record_length+8))
                    else:
                        cleanForm+='12x '

            # removes all unwanted information from the binary structure
            try:
                cleaned = b''.join(struct.unpack(cleanForm, bindata))
            except:
                print("ERROR: I do not recognise the data structure!")
                return None
            del bindata

            # reads the catalog from the cleaned bynary structure
            thiscat = np.frombuffer(cleaned, dtype=stored_dtype)
            del cleaned

            # removes unwanted columns from the catalog
            if self.data is None:
                self.data = np.zeros(NhalosPerFile[myfile], dtype=self.cat_dtype)
            else:
                self.data.resize(self.data.shape[0]+NhalosPerFile[myfile])

            for name in self.data.dtype.names:
                self.data[name][-NhalosPerFile[myfile]:]=thiscat[name]
            del thiscat

            if not silent:
                print("done with file {}".format(myfname))

        if not silent:
            print("Reading catalog done, {} groups found".format(NhalosPerFile.sum()))

        self.Nhalos = len(self.data)

        # Create few pointers to make it compatible with previous versions
        self.Mass  = self.data['Mass']
        self.pos   = self.data['pos']
        if record_length>48:
            self.Npart = self.data['npart']
        self.vel   = self.data['vel']

class plc:

    '''
    Reads a pinocchio catalog on the past light cone
    Usage: myplc=rp.plc(filename, silent=False, first_file=None, last_file=None, onlyNfiles=False)

    filename: name of catalog file, ending with .out. The code will recognise if the catalog is writen in several files
    silent: if True, the script does not give messages in the standard output
    first_file ,last_file: if the catalog if written in several files, it allows to select a limited range of output files
    onlyNfiles: it computes Nfiles and exits returning its value, without reading the catalog

    Returns:
    myplc.data: the catalog, a structured numpy array 
    myplc.cat_dtype: the data type of the catalog:
        [('name', numpy.uint64),
        ('truez', numpy.float32),
        ('pos', numpy.float32, 3),       (absent in light format)
        ('vel', numpy.float32, 3),       (absent in light format)
        ('Mass', numpy.float32),
        ('theta', numpy.float32),
        ('phi', numpy.float32),
        ('vlos', numpy.float32),         (absent in light format)
        ('obsz', numpy.float32)]
    myplc.Nhalos: number of halos in the catalog
    myplc.Nfiles: number of files in which the catalog is written
    
    '''

    def __init__(self,filename,silent=False,first_file=None,last_file=None,onlyNfiles=False):

        if VERBOSE:
            silent=False

        # checks that the filename contains 'plc'
        if not 'plc' in filename:
            print("Are you sure you are providing the right file name?")
            if 'catalog' in filename:
                print("...this looks like a catalog file, please read it with rp.catalog(filename)")
            elif 'histories' in filename:
                print("...this looks like a histories file, please read it with rp.histories(filename)")
            return None

        # checks that the input file ends by ".out"
        last_ext=filename.rfind('.')
        if filename[last_ext:]!='.out':

            print("The catalog file should end with .out, the file number extension will be checked by the code")
            return None

        # checks that the file exists, of that there are multiple files
        if not os.path.exists(filename):

            if not os.path.exists(filename+'.0'):

                print("file {} or {} not found:".format(filename,filename+'.0'))
                return None

            else:

                Nfiles=1
                while os.path.exists(filename+'.{}'.format(Nfiles)):
                    Nfiles+=1
                if not silent:
                    print("The catalog is written in {} files".format(Nfiles))

        else:

            Nfiles=1
            if not silent:
                print("The catalog is written in 1 file")

        self.Nfiles=Nfiles
        if onlyNfiles:
            return

        # opens the (first) file and reads the record length
        if Nfiles==1:
            if not silent:
                print('reading header of file '+filename)
            reading = np.fromfile(filename,dtype=np.int32,count=3)
        else:
            if not silent:
                print('reading header of file '+filename+'.0')
            reading = np.fromfile(filename+'.0',dtype=np.int32,count=3)

        # reads the record length
        if reading[0]==4:
            record_length=reading[1]
            newRun=True
            if not silent:
                if record_length==32:
                    print('This is new light output format, record length: {}'.format(record_length))
                else:
                    print('This is new full output format, record length: {}'.format(record_length))
        else:
            record_length=reading[0]
            newRun=False
            if not silent:
                print('This is classic output format, record length: {}'.format(record_length))
                  
        # sets the record
        if record_length==104:

            self.cat_dtype = [ ( 'name'  , np.uint64 ),
                               ( 'truez' , np.float64 ),
                               ( 'pos'   , np.float64,3 ),
                               ( 'vel' , np.float64,3 ),
                               ( 'Mass' , np.float64 ),
                               ( 'theta' , np.float64 ),
                               ( 'phi' , np.float64 ),
                               ( 'vlos' , np.float64 ),
                               ( 'obsz' , np.float64 ) ]

            stored_dtype   = [ ('fort',   np.int32),
                               ( 'name',  np.uint64 ),
                               ( 'truez', np.float64 ),
                               ( 'pos',   np.float64,3 ),
                               ( 'vel',   np.float64,3 ),
                               ( 'Mass' , np.float64 ),
                               ( 'theta', np.float64 ),
                               ( 'phi',   np.float64 ),
                               ( 'vlos' , np.float64 ),
                               ( 'obsz' , np.float64 ) ,
                               ( 'trof',  np.int32) ]

        elif record_length==56:

            self.cat_dtype = [ ( 'name',  np.uint64 ),
                               ( 'truez', np.float32 ),
                               ( 'pos',   np.float32,3 ),
                               ( 'vel',   np.float32,3 ),
                               ( 'Mass',  np.float32 ),
                               ( 'theta', np.float32 ),
                               ( 'phi',   np.float32 ),
                               ( 'vlos',  np.float32 ),
                               ( 'obsz',  np.float32 ) ]
            if newRun:

                stored_dtype   = [ ( 'name',  np.uint64 ),
                                   ( 'truez', np.float32 ),
                                   ( 'pos',   np.float32,3 ),
                                   ( 'vel',   np.float32,3 ),
                                   ( 'Mass',  np.float32 ),
                                   ( 'theta', np.float32 ),
                                   ( 'phi',   np.float32 ),
                                   ( 'vlos',  np.float32 ),
                                   ( 'obsz',  np.float32 ) ]

            else:

                stored_dtype   = [ ( 'fort',  np.int32),
                                   ( 'name',  np.uint64 ),
                                   ( 'truez', np.float32 ),
                                   ( 'pos',   np.float32,3 ),
                                   ( 'vel',   np.float32,3 ),
                                   ( 'Mass',  np.float32 ),
                                   ( 'theta', np.float32 ),
                                   ( 'phi',   np.float32 ),
                                   ( 'vlos',  np.float32 ),
                                   ( 'obsz',  np.float32 ),
                                   ( 'trof',  np.int32) ]


        elif record_length==32:

            self.cat_dtype = [ ( 'name'  , np.uint64 ),
                               ( 'truez' , np.float32 ),
                               ( 'Mass'  , np.float32 ),
                               ( 'theta' , np.float32 ),
                               ( 'phi'   , np.float32 ),
                               ( 'obsz'  , np.float32 ) ]

            stored_dtype   = [ ( 'name'  , np.uint64 ),
                               ( 'truez' , np.float32 ),
                               ( 'Mass'  , np.float32 ),
                               ( 'theta' , np.float32 ),
                               ( 'phi'   , np.float32 ),
                               ( 'obsz'  , np.float32 ),
                               ( 'pad'   , np.float32 ) ]

        else:
            print("sorry, I do not recognize this record length")
            return None

        # decides what files to read
        if Nfiles>1:
            if first_file is None:
                first_file=0
            elif first_file<0:
                first_file=0
            elif first_file>Nfiles:
                first_file=Nfiles
            if last_file is None:
                last_file=Nfiles
            else:
                last_file += 1      
                if last_file<first_file:
                    last_file=first_file+1
                elif last_file>Nfiles:
                    last_file=Nfiles
            # this is to be used to define a pythonic range
            if not silent:
                print("I will read files in the python range from {} to {}".format(first_file,last_file))
        else:
            first_file=0
            last_file=Nfiles

        # prepares to read the file(s)
        self.data=None
        NhalosPerFile=np.zeros(Nfiles,dtype=np.int64)

        # loops on the files to be read
        for myfile in range(first_file,last_file):

            if Nfiles==1:
                myfname = filename
            else:
                myfname = filename+'.{}'.format(myfile)

            if VERBOSE:
                print("reading file {}".format(myfname))

            # reads the file as a binary object
            with open(myfname,'rb') as f:
                bindata=f.read()
                f.close()
            filesize = len(bindata)
            
            if newRun:

                # scrolls the file to map the blocks

                if VERBOSE:
                    print("scrolling file to reconstruct the blocks:")
                cleanForm='12x '
                pos=12
                vechalo=[]
                Nblocks=0
                while pos < filesize:
                    # reads the number of halos to read
                    vec=struct.Struct('iii').unpack(bindata[pos:pos+12])[1]
                    pos+=12
                    vechalo.append(vec)
                    if vec==0:
                        print("THIS SHOULD NOT HAPPEN!")
                        return None
                    Nblocks+=1
                    pos += 8+vec*record_length
                    cleanForm+='16x {}s 4x '.format(vec*record_length)

                vechalo = np.asarray(vechalo)
                NhalosPerFile[myfile]=vechalo.sum()

                if VERBOSE:
                    print(f"data are written in {Nblocks} blocks")

                # removes all unwanted information from the binary structure
                try:
                    cleaned = b''.join(struct.unpack(cleanForm, bindata))
                    if VERBOSE:
                        print("cleaning of binary object done")
                except:
                    print("ERROR: I do not recognise the data structure!")
                    return None
                del bindata

                # reads the catalog from the cleaned bynary structure
                thiscat = np.frombuffer(cleaned, dtype=stored_dtype)
                if VERBOSE:
                    print("catalog extracted")
                del cleaned

            else:

                # reads the catalog directly from the binary object
                thiscat = np.frombuffer(bindata, dtype=stored_dtype)
                if VERBOSE:
                    print("catalog extracted")
                del bindata
                NhalosPerFile[myfile]=len(thiscat)

            if newRun:
                file_length = 12 + Nblocks*20 + NhalosPerFile[myfile]*record_length
            else:
                file_length = NhalosPerFile[myfile]*(record_length+8)
            if file_length != filesize:
                print(f'ERROR: inconsistency in the file size, should be {file_length} but I found {filesize} bytes')
                return None
            elif VERBOSE:
                print(f'predicted file length {file_length} matches with file size {filesize}')


            # removes unwanted columns from the catalog
            if self.data is None:
                self.data = np.zeros(NhalosPerFile[myfile], dtype=self.cat_dtype)
            else:
                self.data.resize(self.data.shape[0]+NhalosPerFile[myfile])

            for name in self.data.dtype.names:
                self.data[name][-NhalosPerFile[myfile]:]=thiscat[name]
            del thiscat


            if not silent:
                print("done with file {}".format(myfname))

        self.Nhalos = len(self.data)

        if not silent:
            print("Reading plc done, {} groups found".format(NhalosPerFile.sum()))



class histories:
    '''
    Reads a pinocchio histories catalog
    Usage: myhist=rp.histories(filename, silent=False, first_file=None, last_file=None)

    filename: name of catalog file, ending with .out. The code will recognise if the catalog is writen in several files
    silent: if True, the script does not give messages in the standard output
    first_file ,last_file: if the catalog if written in several files, it allows to select a limited range of output files

    Returns:
    myhist.data: the catalog, a structured numpy array
    myhist.cat_dtype: the data type of the catalog:
       [('name', numpy.uint64),
        ('nickname', numpy.int32),
        ('link', numpy.int32),
        ('merged_with', numpy.int32),
        ('mass_at_merger', numpy.int32),
        ('mass_of_main', numpy.int32),
        ('z_merging', numpy.float32),
        ('z_peak', numpy.float32),
        ('z_appear', numpy.float32)]
    myhist.Nfiles: number of files in which the catalog is written
    myhist.Ntrees: number of trees in the catalog
    myhist.Nbranches_tot: total number of branches in the catalog
    myhist.Nbranches: array of size Ntrees giving the number of branches for each tree
    myhist.pointers; array of size Ntrees giving a pointer to the start of the tree in the catalog

    '''

    def __init__(self,filename,silent=False,first_file=None,last_file=None):

        if VERBOSE:
            silent=False

        # checks that the filename contains 'catalog'
        if not 'histories' in filename:
            print("Are you sure you are providing the right file name?")
            if 'plc' in filename:
                print("...this looks like a plc file, please read it with rp.plc(filename)")
            elif 'catalog' in filename:
                print("...this looks like a catalog file, please read it with rp.catalog(filename)")
            return None

        # checks that the input file ends by ".out"
        last_ext=filename.rfind('.')
        if filename[last_ext:]!='.out':

            print("The history file should end with .out, the file number extension will be checked by the code")
            return None

        # checks that the file exists, of that there are multiple files
        if not os.path.exists(filename):

            if not os.path.exists(filename+'.0'):

                print("file {} or {} not found:".format(filename,filename+'.0'))
                return None

            else:

                Nfiles=1
                while os.path.exists(filename+'.{}'.format(Nfiles)):
                    Nfiles+=1
                if not silent:
                    print("The catalog is written in {} files".format(Nfiles))

        else:

            Nfiles=1
            if not silent:
                print("The catalog is written in 1 file")

        # opens the (first) file and reads the record length
        if Nfiles==1:
            if not silent:
                print('opening file '+filename)
            reading = np.fromfile(filename,dtype=np.int32,count=12)
            FileLength = os.path.getsize(filename)
        else:
            if not silent:
                print('opening file '+filename+'.0')
            reading = np.fromfile(filename+'.0',dtype=np.int32,count=12)
            FileLength = os.path.getsize(filename+'.0')

        # reads the header and the record length
        if reading[1]>10:
            newRun=True
            record_length=reading[1]
            Nslices=1
            Light = FileLength==np.int64(record_length)*np.int64(reading[5])+np.int64(4*7)
            if not silent:
                if not Light:
                    print('This is new output format, record length: {}'.format(record_length))
                else:
                    print('This is new light output format, record length: {}'.format(record_length))
        else:

            if not silent:
                print("WARNING: reading of V4 histories is very slow")

            newRun=False
            record_length=reading[11]
            Nslices=reading[1]
            if not silent:
                print('This is classic output format, record length: {}'.format(record_length))
                if Nslices==1:
                    print('The box has been fragmented in 1 slice')
                else:
                    print('The box has been fragmented in {} slices'.format(Nslices))


        if record_length==40:

            self.cat_dtype=[ ( 'name'           , np.uint64 ),
                             ( 'nickname'       , np.int32 ),
                             ( 'link'           , np.int32 ),
                             ( 'merged_with'    , np.int32 ),
                             ( 'mass_at_merger' , np.int32 ),
                             ( 'mass_of_main'   , np.int32 ),
                             ( 'z_merging'      , np.float32 ),
                             ( 'z_peak'         , np.float32 ),
                             ( 'z_appear'       , np.float32 ) ]

        elif record_length==56:

            self.cat_dtype=[ ( 'name'           , np.uint64 ),
                             ( 'nickname'       , np.int32 ),
                             ( 'link'           , np.int32 ),
                             ( 'merged_with'    , np.int32 ),
                             ( 'mass_at_merger' , np.int32 ),
                             ( 'mass_of_main'   , np.int32 ),
                             ( 'z_merging'      , np.float64 ),
                             ( 'z_peak'         , np.float64 ),
                             ( 'z_appear'       , np.float64 ) ]

            stored_dtype  =[ ( 'fort'           , np.int32 ),
                             ( 'name'           , np.uint64 ),
                             ( 'nickname'       , np.int32 ),
                             ( 'link'           , np.int32 ),
                             ( 'merged_with'    , np.int32 ),
                             ( 'mass_at_merger' , np.int32 ),
                             ( 'mass_of_main'   , np.int32 ),
                             ( 'pad'            , np.int32 ),
                             ( 'z_merging'      , np.float64 ),
                             ( 'z_peak'         , np.float64 ),
                             ( 'z_appear'       , np.float64 ),
                             ( 'trof'           , np.int32 ) ]

        else:
            print("sorry, I do not recognize this record length")
            return None

        self.Nfiles=Nfiles
        # decides what files to read
        if Nfiles>1:
            if first_file is None:
                first_file=0
            elif first_file<0:
                first_file=0
            elif first_file>Nfiles:
                first_file=Nfiles
            if last_file is None:
                last_file=Nfiles
            else:
                last_file += 1      
                if last_file<first_file:
                    last_file=first_file+1
                elif last_file>Nfiles:
                    last_file=Nfiles
            # this is to be used to define a pythonic range
            if not silent:
                print("I will read files in the python range from {} to {}".format(first_file,last_file))
        else:
            first_file=0
            last_file=Nfiles

        # prepares to read the file(s)
        self.data=None
        self.Nbranches=None
        TTotal=np.int64(0)
        BTotal=np.int64(0)

        for myfile in range(first_file,last_file):

            if Nfiles==1:
                myfname = filename
            else:
                myfname = filename+'.{}'.format(myfile)

            if VERBOSE:
                print("reading file {}".format(myfname))

            # reads the file as a binary object
            with open(myfname,'rb') as f:
                bindata=f.read()
                f.close()
            FileLength = len(bindata)
            Tthisfile, Bthisfile = struct.Struct('ii').unpack(bindata[16:24])
            TTotal += Tthisfile
            BTotal += Bthisfile

            if newRun:

                if Light:
                    
                    # light format is straightforward to read
                    if VERBOSE:
                        print("reading in data")

                    if self.data is None:
                        self.data = np.copy(np.frombuffer(bindata[28:], dtype=self.cat_dtype))
                    else:
                        thiscat = np.frombuffer(bindata[28:], dtype=self.cat_dtype)
                        self.data.resize(self.data.shape[0]+len(thiscat))
                        self.data[-len(thiscat):]=np.copy(thiscat)
                        del thiscat

                else:

                    # reading standard V5 format
                    cleanForm='28x '
                    pos=28
                    Nblocks=0
                    while pos < FileLength:
                        # reads the number of branches
                        Nthisblock = struct.Struct('iii').unpack(bindata[pos:pos+12])[1]
                        pos+=16
                        Nbranches = np.frombuffer(bindata[pos:pos+Nthisblock*4],
                                                  dtype=np.int32,count=Nthisblock)
                        Bthisblock=Nbranches.sum()
                        pos+=Nthisblock*4 + 8 + Bthisblock*record_length + 4
                        cleanForm+='{}x {}s 4x'.format(12+Nthisblock*4+12,Bthisblock*record_length)
                        Nblocks+=1

                        if self.Nbranches is None:
                            self.Nbranches = np.copy(Nbranches)
                        else:
                            self.Nbranches.resize(self.Nbranches.shape[0]+Nthisblock)
                            self.Nbranches[-Nthisblock:]=np.copy(Nbranches)
                        del Nbranches

                    if VERBOSE:
                        print(f"data are written in {Nblocks} blocks")

                    # removes all unwanted information from the binary structure
                    try:
                        cleaned = b''.join(struct.unpack(cleanForm, bindata))
                        if VERBOSE:
                            print("cleaning of binary object done")
                    except:
                        print("ERROR: I do not recognise the data structure!")
                        return None
                    del bindata

                    thiscat = np.frombuffer(cleaned, dtype=self.cat_dtype)
                    if self.data is None:
                        self.data = np.copy(thiscat)
                    else:
                        self.data.resize(self.data.shape[0]+len(thiscat))
                        self.data[-len(thiscat):]=np.copy(thiscat)
                    del thiscat

            else:

                # first reads the total number of branches
                pos=12
                Bthisfile=0
                Tthisfile=0
                while pos < FileLength:
                    Nt, Nb = struct.Struct('iiii').unpack(bindata[pos:pos+16])[1:3]
                    pos+=16 + Nt*16 + Nb*(record_length+8)
                    Tthisfile+=Nt
                    Bthisfile+=Nb

                if self.data is None:
                    cNb=0
                    cTr=0
                    self.data      = np.empty(Bthisfile, dtype=self.cat_dtype)
                    self.Nbranches = np.empty(Tthisfile, dtype=np.int32)
                else:
                    cNb=self.Nbranches.shape[0]
                    cTr=self.data.shape[0]
                    self.data.resize(self.data.shape[0]+Bthisfile)
                    self.Nbranches.resize(self.Nbranches.shape[0]+Tthisfile)

                pos=12
                while pos < FileLength:
                    Nt = struct.Struct('iiii').unpack(bindata[pos:pos+16])[1]
                    pos+=16
                    for t in range(Nt):
                        Nb = struct.Struct('iiii').unpack(bindata[pos:pos+16])[2]
                        pos+=16
                        self.Nbranches[cNb]=Nb
                        cNb+=1
                        thiscat=np.frombuffer(bindata[pos:pos+Nb*(record_length+8)],dtype=stored_dtype,count=Nb)
                        for name in self.data.dtype.names:
                            self.data[name][cTr:cTr+Nb]=thiscat[name]
                        del thiscat
                        pos+=Nb*(record_length+8)
                        cTr+=Nb


            if newRun:
                if Light:
                    file_length = 28 + Bthisfile*record_length
                else:
                    file_length = 28 + Nblocks*28 + Tthisfile*4 + Bthisfile*record_length
            else:
                file_length = 12 + Nslices*16 + Tthisfile*16 + Bthisfile*(record_length+8)
            if FileLength != file_length:
                print("Error: inconsistent length for the file {}, I expected {} and found {}".format(myfname,FileLength,file_length))
                return None
            elif VERBOSE:
                print("File size of {} is as expected".format(myfname))


        if newRun and not Light:
            # cumulate pointers
            self.pointers = np.cumsum(np.insert(self.Nbranches,0,0))[:TTotal]
        else:
            if not silent:
                print("Building pointers...")
            po = np.empty(TTotal+1, dtype=np.int32)
            self.Nbranches = np.empty(TTotal, dtype=np.int32)

            po[0]=0
            for i in range(TTotal):
                self.Nbranches[i]=self.data[po[i]]['nickname']
                po[i+1]=po[i]+self.Nbranches[i]

            self.pointers = po[0:TTotal]
            del po
            
        self.Ntrees=TTotal
        self.Nbranches_tot=BTotal

        if not silent:
            print("Reading catalog done")

